'''
    This section of the codebase is used for construction of route paths using the 
    environmental mesh between a series of user defined waypoints
'''

import copy, json, ast, warnings
import numpy as np
import pandas as pd
from tqdm import tqdm
from shapely import wkt
from shapely.geometry.polygon import Point
import geopandas as gpd
import logging

from pandas.core.common import SettingWithCopyWarning
warnings.simplefilter(action="ignore", category=SettingWithCopyWarning)

from polar_route.crossing import NewtonianDistance, NewtonianCurve


def _flattenCases(id,mesh):
    neighbour_case = []
    neighbour_indx = []
    neighbours = mesh['neighbour_graph'][id]
    for case in neighbours.keys():
        for neighbour in neighbours[case]:
            neighbour_case.append(int(case))
            neighbour_indx.append(int(neighbour))
    return neighbour_case, neighbour_indx


class RoutePlanner:
    """
        ---

        RoutePlanner optimises the route paths between a series of waypoints. 
        The routes are constructed in a two stage process:

        compute_routes: uses a mesh based Dijkstra method to determine the optimal routes 
                        between a series of waypoint.

        compute_smoothed_routes: smooths the compute_routes using information from the environmental mesh
                                to determine mesh independent optimal route paths


        ---

        Attributes:
            waypoints (dict): A dictionary of the waypoints supplied by the user in the region
                of interest. The waypoints are of the form:

               {\n
                    "Name":{\n
                        '0':"Falklands",\n
                        '1':"Rothera",\n
                        ...\n
                    },\n
                    "Lat":{\n
                        '0':-52.6347222222,
                        '1':-75.26722,\n
                        ...\n
                    },\n
                    "Long":{\n
                        ...\n
                    },\n
                    "index":{\n
                        ...\n
                    }\n
               }



            paths (geojson): A GeoJSON of all paths constructed. The paths are in the form:

                {\n
                    'types':'FeatureCollection',\n
                    "features":{[\n
                        'type':'feature',\n
                        'geometry':{\n
                            'type': 'LineString',

                            'coordinates': [[-27.21694, -75.26722],\n
                                            [-27.5, -75.07960297382266],\n
                                            [-27.619238882768894, -75.0],\n
                                            ...]\n
                        },
                        'properties':{\n
                            'from': 'Halley',\n
                            'to': 'Rothera',\n
                            'traveltime': [0.0,\n
                                        0.03531938671648596,\n
                                        0.050310986633880575,\n
                                        ...],\n
                            'fuel': [0.0,\n
                                    0.9648858923588642,\n
                                    1.3745886107069096,\n
                                    ...],\n
                            'times': ['2017-01-01 00:00:00',
                                    '2017-01-01 00:50:51.595036800',
                                    '2017-01-01 01:12:26.869276800',
                                    ...]\n
                        }\n
                    ]}\n
                }\n

        ---

    """

    def __init__(self, mesh, waypoints, cost_func=NewtonianDistance):

        """
            Constructs the routes from information given in the config file.

            Args:
                config (dict): config file which defines the attributes required for the route construction. 
                    Sections required for the route construction are as follows\n
                    \n
                    {\n
                        "config": {...\n
                            "Route_Info":{\n
                                    "objective_function": (string) currently either 'traveltime' or 'fuel',\n
                                    "path_variables": list of (string),\n
                                    "waypoints_path": (string),\n
                                    "source_waypoints": list of (string),\n
                                    "end_waypoints": list of (string),\n
                                    "vector_names": (list of (string),\n
                                    "zero_currents": (boolean),\n
                                    "variable_speed" (boolean),\n
                                    "time_unit" (string),\n
                                    "early_stopping_criterion" (boolean),\n
                                    "save_dijkstra_graphs": (boolean),\n
                                    "smooth_path":{\n
                                        "max_iteration_number":(int),\n
                                        "minimum_difference":(float),\n
                                    }\n
                                },\n
                        }\n
                    }\n

                cost_func (func): Crossing point cost function for Dijkstra Path creation. For development purposes only !
        """

        # Load in the current cell structure & Optimisation Info̦
        # self.mesh    = copy.copy(mesh)
        self.mesh = mesh
        self.config  = self.mesh['config']

        waypoints_df = pd.read_csv(waypoints)
        source_waypoints_df   = waypoints_df[waypoints_df['Source'] == "X"]
        des_waypoints_df      = waypoints_df[waypoints_df['Destination'] == "X"]

        self.source_waypoints = list(source_waypoints_df['Name'])
        self.end_waypoints    = list(des_waypoints_df['Name'])

        # Creating a blank path construct
        self.paths          = None
        self.smoothed_paths = None
        self.dijkstra_info = {}

        # ====== Loading Mesh & Neighbour Graph ======
        # Formatting the Mesh and Neighbour Graph to the right form
        self.neighbour_graph = pd.DataFrame(self.mesh['cellboxes']).set_index('id')
        self.neighbour_graph['geometry'] = self.neighbour_graph['geometry'].apply(wkt.loads)
        self.neighbour_graph = gpd.GeoDataFrame(self.neighbour_graph, crs='EPSG:4326', geometry='geometry')

        # Removing any point not in accessible positions
        self.neighbour_graph = self.neighbour_graph.loc[list(self.mesh['neighbour_graph'].keys())]

        # Using neighbour graph to determine neighbour indices and cases
        self.neighbour_graph['case'], self.neighbour_graph['neighbourIndex'] = zip(*self.neighbour_graph.apply(lambda row: _flattenCases(row.name, self.mesh), axis=1))

        self.neighbour_graph.index = self.neighbour_graph.index.astype(int)

        # Renaming the vector columns
        self.neighbour_graph = self.neighbour_graph.rename(columns={self.config['Route_Info']['vector_names'][0]: "Vector_x", 
                                                                    self.config['Route_Info']['vector_names'][1]: "Vector_y"})

        # ====== Speed Function Checking ======
        # Checking if Speed defined in file
        if 'Speed' not in self.neighbour_graph:
            self.neighbour_graph['Speed'] = self.config["Vessel"]["Speed"]

        # ====== Objective Function Information ======
        #  Checking if objective function is in the cellgrid            
        if self.config['Route_Info']['objective_function'] != 'traveltime':
            if self.config['Route_Info']['objective_function'] not in self.neighbour_graph:
                raise Exception("Objective Function require '{}' column in mesh dataframe".format(self.config['Route_Info']['objective_function']))

        # ===== Dijkstra Graph =====
        # Adding the required columns needed for the dijkstra graph
        self.neighbour_graph['positionLocked']          = False
        for vrbl in self.config['Route_Info']['path_variables']:
            self.neighbour_graph['shortest_{}'.format(vrbl)]    = np.inf
        self.neighbour_graph['neighbourTravelLegs']     = [list() for x in range(len(self.neighbour_graph.index))]
        self.neighbour_graph['neighbourCrossingPoints'] = [list() for x in range(len(self.neighbour_graph.index))]
        self.neighbour_graph['pathIndex']               = [list() for x in range(len(self.neighbour_graph.index))]
        for vrbl in self.config['Route_Info']['path_variables']:
            self.neighbour_graph['path_{}'.format(vrbl)] = [list() for x in range(len(self.neighbour_graph.index))]
        self.neighbour_graph['pathPoints']               = [list() for x in range(len(self.neighbour_graph.index))]

        # ====== Defining the cost function ======
        self.cost_func       = cost_func

        # ====== Outlining some constant values ======
        self.unit_shipspeed = self.config['Vessel']['Unit']
        self.unit_time      = self.config['Route_Info']['time_unit']
        self.zero_currents  = self.config['Route_Info']['zero_currents']
        self.variable_speed  =self.config['Route_Info']['variable_speed']

        if not self.variable_speed:
            self.neighbour_graph['Speed'] = self.config["Vessel"]["Speed"]

        # ====== Waypoints ======
        self.mesh['waypoints'] = waypoints_df

        # Dropping waypoints outside domain
        self.mesh['waypoints'] = self.mesh['waypoints'][\
                                                              (self.mesh['waypoints']['Long'] >= self.config['Mesh_info']['Region']['longMin']) &\
                                                              (self.mesh['waypoints']['Long'] <=  self.config['Mesh_info']['Region']['longMax']) &\
                                                              (self.mesh['waypoints']['Lat'] <=  self.config['Mesh_info']['Region']['latMax']) &\
                                                              (self.mesh['waypoints']['Lat'] >=  self.config['Mesh_info']['Region']['latMin'])] 

        # # Initialising Waypoints positions and cell index
        wpts = self.mesh['waypoints']
        wpts['index'] = np.nan
        for idx,wpt in wpts.iterrows():
            indices = self.neighbour_graph[self.neighbour_graph['geometry'].contains(Point(wpt[['Long','Lat']]))].index
            # Waypoint is not within a mesh cell, but could still be on the edge of one. So perturbing the position slightly to the north-east and checking again. 
            #If this is not the case then the waypoint is not within the navitagatable domain and will continue
            if len(indices) == 0:
                try:
                    indices = mesh[(mesh['geometry'].contains(Point(wpt[['Long','Lat']]+1e-5)))].index
                except:
                    continue
            if len(indices) == 0:
                continue
            if len(indices) > 1:
                raise Exception('Wapoint lies in multiple cell boxes. Please check mesh ! ')
            else:
                wpts['index'].loc[idx] = int(indices[0])

        self.mesh['waypoints'] = wpts[~wpts['index'].isnull()]
        self.mesh['waypoints']['index'] = self.mesh['waypoints']['index'].astype(int)
        self.mesh['waypoints'] =  self.mesh['waypoints'].to_json()

        # ==== Printing Configuration and Information
        self.mesh['waypoints'] =  pd.read_json(self.mesh['waypoints'])


    def to_json(self):
        '''
            Outputing the information in JSON format
        '''
        mesh = copy.copy(self.mesh)
        mesh['waypoints'] = mesh['waypoints'].to_dict()
        output_json = json.loads(json.dumps(mesh))
        del mesh
        return output_json

    def _dijkstra_paths(self, start_waypoints, end_waypoints):
        """
            FILL
        """

        geojson = dict()
        geojson['type'] = "FeatureCollection"

        paths = []
        wpts_s = self.mesh['waypoints'][self.mesh['waypoints']['Name'].isin(start_waypoints)]
        wpts_e = self.mesh['waypoints'][self.mesh['waypoints']['Name'].isin(end_waypoints)]
        for _, wpt_a in wpts_s.iterrows():
            wpt_a_name  = wpt_a['Name']
            wpt_a_index = int(wpt_a['index'])
            wpt_a_loc   = [[wpt_a['Long'],wpt_a['Lat']]]
            for _, wpt_b in wpts_e.iterrows():
                wpt_b_name  = wpt_b['Name']
                wpt_b_index = int(wpt_b['index'])
                wpt_b_loc   = [[wpt_b['Long'],wpt_b['Lat']]]
                if not wpt_a_name == wpt_b_name:
                    try:
                        graph = self.dijkstra_info[wpt_a_name]
                        path = dict()
                        path['type'] = "Feature"
                        path['geometry'] = {}
                        path['geometry']['type'] = "LineString"
                        path_points = (np.array(wpt_a_loc+list(np.array(graph['pathPoints'].loc[wpt_b_index])[:-1, :])+wpt_b_loc))
                        path['geometry']['coordinates'] = path_points.tolist()

                        path['properties'] = {}
                        path['properties']['name'] = 'Route Path - {} to {}'.format(wpt_a_name, wpt_b_name)
                        path['properties']['from'] = '{}'.format(wpt_a_name)
                        path['properties']['to'] = '{}'.format(wpt_b_name)

                        cellIndices  = np.array(graph['pathIndex'].loc[wpt_b_index])
                        path_indices = np.array([cellIndices[0]] + list(np.repeat(cellIndices[1:-1], 2)) + [cellIndices[-1]])
                        path['properties']['CellIndices'] = path_indices.tolist()

                        # Applying in-cell correction for travel-time
                        cost_func    = self.cost_func(source_graph=self.dijkstra_info[wpt_a_name].loc[path_indices[0]],
                                                      neighbour_graph=self.dijkstra_info[wpt_a_name].loc[path_indices[0]],
                                                      unit_shipspeed='km/hr', unit_time=self.unit_time, zerocurrents=self.zero_currents)
                        tt_start = cost_func.waypoint_correction(path_points[0, :], path_points[1, :])
                        cost_func    = self.cost_func(source_graph=self.dijkstra_info[wpt_a_name].loc[path_indices[-1]],
                                                      neighbour_graph=self.dijkstra_info[wpt_a_name].loc[path_indices[0]],
                                                      unit_shipspeed='km/hr', unit_time=self.unit_time, zerocurrents=self.zero_currents)
                        tt_end = cost_func.waypoint_correction(path_points[-1, :], path_points[-2, :])
                        path['properties']['traveltime']     = np.array(graph['path_traveltime'].loc[wpt_b_index])
                        path['properties']['traveltime']     = (path['properties']['traveltime'] - path['properties']['traveltime'][0]) + tt_start
                        path['properties']['traveltime'][-1] = (path['properties']['traveltime'][-2] + tt_end)

                        for vrbl in self.config['Route_Info']['path_variables']:
                            if vrbl == 'traveltime':
                                continue
                            path['properties'][vrbl] = np.cumsum(np.r_[path['properties']['traveltime'][0], np.diff(path['properties']['traveltime'])]*self.dijkstra_info[wpt_a_name].loc[path_indices,'{}'.format(vrbl)].to_numpy()).tolist()
                        path['properties']['traveltime'] = path['properties']['traveltime'].tolist()
                        paths.append(path)

                    except:
                        logging.warning('{} to {} - Failured to construct path direct in the dijkstra information'.format(wpt_a_name,wpt_b_name))

        geojson['features'] = paths
        return geojson

    def _objective_value(self, variable, source_graph, neighbour_graph, traveltime):
        """
            FILL
        """
        if variable == 'traveltime':
            return np.array([source_graph['shortest_traveltime'] + traveltime[0],source_graph['shortest_traveltime'] + np.sum(traveltime)])
        else:
            return np.array([source_graph['shortest_{}'.format(variable)] +\
                    traveltime[0]*source_graph['{}'.format(variable)],
                    source_graph['shortest_{}'.format(variable)] +\
                    traveltime[0]*source_graph['{}'.format(variable)] +\
                    traveltime[1]*neighbour_graph['{}'.format(variable)]])

    def _neighbour_cost(self, wpt_name, minimum_objective_index):
        """
            FILL
        """
        # Determining the nearest neighbour index for the cell
        source_graph   = self.dijkstra_info[wpt_name].loc[minimum_objective_index]

        # Looping over idx
        for idx in range(len(source_graph['case'])):
            indx = source_graph['neighbourIndex'][idx]

            neighbour_graph = self.dijkstra_info[wpt_name].loc[indx]
            case = source_graph['case'][idx]

            # Applying Newton curve to determine crossing point
            cost_func    = self.cost_func(source_graph=source_graph, neighbour_graph=neighbour_graph, case=case,
                                          unit_shipspeed='km/hr', unit_time=self.unit_time,
                                          zerocurrents=self.zero_currents)
            # Updating the Dijkstra graph with the new information
            traveltime, crossing_points,cell_points = cost_func.value()

            source_graph['neighbourTravelLegs'].append(traveltime)
            source_graph['neighbourCrossingPoints'].append(np.array(crossing_points))

            # Using neighbourhood cost determine objective function value
            value = self._objective_value(self.config['Route_Info']['objective_function'], source_graph,
                                          neighbour_graph, traveltime)
            if value[1] < neighbour_graph['shortest_{}'.format(self.config['Route_Info']['objective_function'])]:
                for vrbl in self.config['Route_Info']['path_variables']:
                    value = self._objective_value(vrbl, source_graph, neighbour_graph,traveltime)
                    neighbour_graph['shortest_{}'.format(vrbl)] = value[1]
                    neighbour_graph['path_{}'.format(vrbl)]   = source_graph['path_{}'.format(vrbl)] + list(value)
                neighbour_graph['pathIndex']  = source_graph['pathIndex']  + [indx]
                neighbour_graph['pathPoints'] = source_graph['pathPoints'] + [list(crossing_points)] + [list(cell_points)]
                self.dijkstra_info[wpt_name].loc[indx] = neighbour_graph

        self.dijkstra_info[wpt_name].loc[minimum_objective_index] = source_graph

    def _dijkstra(self, wpt_name):
        """
            FILL
        """
        # Including only the End Waypoints defined by the user
        wpts = self.mesh['waypoints'][self.mesh['waypoints']['Name'].isin(self.end_waypoints)]
        
        # Initialising zero traveltime at the source location
        source_index = int(self.mesh['waypoints'][self.mesh['waypoints']['Name'] == wpt_name]['index'])

        for vrbl in self.config['Route_Info']['path_variables']:
            self.dijkstra_info[wpt_name].loc[source_index, 'shortest_{}'.format(vrbl)] = 0.0
        self.dijkstra_info[wpt_name].loc[source_index, 'pathIndex'].append(source_index)
        
        # # Updating Dijkstra as long as all the waypoints are not visited or for full graph
        if self.config['Route_Info']['early_stopping_criterion']:
            stopping_criterion_indices = wpts['index']
        else:
            stopping_criterion_indices = self.dijkstra_info[wpt_name].index

        while (self.dijkstra_info[wpt_name].loc[stopping_criterion_indices, 'positionLocked'] == False).any():

            # Determining the index of the minimum objective function that has not been visited
            minimum_objective_index = self.dijkstra_info[wpt_name][self.dijkstra_info[wpt_name]['positionLocked'] == False]['shortest_{}'.format(self.config['Route_Info']['objective_function'])].idxmin()
  
            # Finding the cost of the nearest neighbours from the source cell (Sc)
            self._neighbour_cost(wpt_name,minimum_objective_index)

            # Updating Position to be locked
            self.dijkstra_info[wpt_name].loc[minimum_objective_index, 'positionLocked'] = True

    def compute_routes(self):
        """
            Computes the Dijkstra Paths between waypoints. 
            `waypoints` and `paths` are appended to output JSON
        """

        # Subsetting the waypoints
        if len(self.source_waypoints) == 0:
            raise Exception('No source waypoints defined that are accessible')
        if len(self.end_waypoints) == 0:
            raise Exception('No destination waypoints defined that are accessible')

        # Initialising the Dijkstra Info Dictionary
        for wpt in self.source_waypoints:
            self.dijkstra_info[wpt] = copy.copy(self.neighbour_graph)

        # 
        logging.info('============= Dijkstr Path Creation ============\n')

        for wpt in self.source_waypoints:
            logging.info('--- Processing Waypoint = {} ---'.format(wpt))
            if len(self.mesh['waypoints'][self.mesh['waypoints']['Name'] == wpt]) == 0:
                logging.warning('{} not in accessible waypoints, continuing'.format(wpt))
                continue
            elif (len(self.mesh['waypoints']['Name'] == wpt) > 0) and (len(self.mesh['waypoints']['Name'] != wpt) == 0):
                logging.warning('{} is accessible but has no destination waypoints, continuing'.format(wpt))
                continue
            else:
                self._dijkstra(wpt)


        # Using Dijkstra Graph compute path and meta information to all end_waypoints
        self.paths = self._dijkstra_paths(self.source_waypoints, self.end_waypoints)
        self.mesh['paths'] = self.paths
        for ii in range(len(self.mesh['paths']['features'])):
            self.mesh['paths']['features'][ii]['properties']['times'] = [str(ii) for ii in (pd.to_datetime(self.mesh['config']['Mesh_info']['Region']['startTime']) + pd.to_timedelta(self.mesh['paths']['features'][ii]['properties']['traveltime'],unit='days'))]

    def compute_smoothed_routes(self):
        """
            Using the previously constructed Dijkstra paths smooth the paths to remove mesh features 
            `paths` will be updated in the output JSON
        """
        maxiter     = self.config['Route_Info']['smooth_path']['max_iteration_number']
        minimumDiff = self.config['Route_Info']['smooth_path']['minimum_difference']

        # 
        SmoothedPaths = []
        geojson = {}
        geojson['type'] = "FeatureCollection"

        if len(self.paths['features']) == 0:
            logging.warning('Paths not constructed as there was no dijkstra paths created')
            return
        Pths = copy.deepcopy(self.paths)['features']  

        logging.info('========= Determining Smoothed Paths ===========\n')

        for ii in range(len(Pths)):
                Path = Pths[ii]
                counter = 0 

                logging.info('---Smoothing {}'.format(Path['properties']['name']))

                nc          = NewtonianCurve(self.dijkstra_info[Path['properties']['from']], self.config, maxiter=1,
                                             zerocurrents=self.zero_currents)
                nc.pathIter = maxiter

                org_path_points = np.array(Path['geometry']['coordinates'])
                org_cellindices = np.array(Path['properties']['CellIndices'])

                # -- Generating a dataframe of the case information -- 
                Points      = np.concatenate([org_path_points[0,:][None,:],org_path_points[1:-1:2],org_path_points[-1,:][None,:]])
                cellIndices = np.concatenate([[org_cellindices[0]],[org_cellindices[0]],org_cellindices[1:-1:2],[org_cellindices[-1]],[org_cellindices[-1]]])
                cellDijk    = [nc.neighbour_graph.loc[ii] for ii in cellIndices]
                nc.CrossingDF  = pd.DataFrame({'cx':Points[:,0],'cy':Points[:,1],'cellStart':cellDijk[:-1],'cellEnd':cellDijk[1:]})

                # -- Determining the cases from the cell information. If within cell then case 0 -- 
                Cases = []
                for idx, row in nc.CrossingDF.iterrows():
                    try:
                        Cases.append(row.cellStart['case'][np.where(np.array(row.cellStart['neighbourIndex']) == row.cellEnd.name)[0][0]])
                    except:
                        Cases.append(0)
                nc.CrossingDF['case'] = Cases
                nc.CrossingDF.index = np.arange(int(nc.CrossingDF.index.min()), int(nc.CrossingDF.index.max()*1e3 + 1e3), int(1e3))

                nc.orgDF = copy.deepcopy(nc.CrossingDF)
                iter = 0
                # try:

                # while iter < nc.pathIter:
                pbar = tqdm(np.arange(nc.pathIter))

                # Determining the computational time averaged across all pairs
                self.allDist = []
                self.allDist2 = []

                self.nc = nc

                for iter in pbar:
                    nc.previousDF = copy.deepcopy(nc.CrossingDF)
                    id = 0
                    while id <= (len(nc.CrossingDF) - 3):
                        #try:
                        nc.triplet = nc.CrossingDF.iloc[id:id+3]
                
                        nc._updateCrossingPoint()
                        self.nc = nc

                        # -- Horseshoe Case Detection -- 
                        nc._horseshoe()

                        # -- Removing reseverse cases                    
                        nc._reverseCase()
                        # except:
                        #     break

                        id += 1+nc.id

                    if  id <= (len(nc.CrossingDF) - 3):
                        print('Path Smoothing Failed!')
                        break


                    self.nc = nc

                    nc._mergePoint()
                    self.nc = nc
                    iter+=1
                        
                    # Stop optimisation if the points are within some minimum difference
                    if len(nc.previousDF) == len(nc.CrossingDF):
                        Dist = np.mean(np.sqrt((nc.previousDF['cx'].astype(float) - nc.CrossingDF['cx'].astype(float))**2 + (nc.previousDF['cy'].astype(float) - nc.CrossingDF['cy'].astype(float))**2))
                        self.allDist.append(Dist)
                        pbar.set_description("Mean Difference = {}".format(Dist))

                        if (Dist==np.min(self.allDist)) and len(np.where(abs(self.allDist - np.min(self.allDist)) < 1e-3)[0]) > 500:
                            logging.info('{} iterations - dDist={}  - early stopping terminated oscilation, returning lowest misfit path - Type 1'.format(iter,Dist))
                            break
                        if (Dist < minimumDiff) and (Dist != 0.0):
                            logging.info('{} iterations - dDist={}'.format(iter, Dist))
                            break
                    # else:
                    #     if 'Dist' in locals():
                    #         self.allDist2.append(Dist)
                    #         if (np.sum((np.array(self.allDist2) - Dist)[-5:]) < 1e-6) and (iter>50) and len(self.allDist2)>50:
                    #             if self.verbose:
                    #                 print('{} iterations - dDist={}  - early stopping terminated oscilation, returning lowest misfit path - Type 2'.format(iter,Dist))
                    #             break




                # Determining the traveltime 
                TravelTimeLegs,DistanceLegs,pathIndex = nc.objective_function()
                FuelLegs  = TravelTimeLegs*self.neighbour_graph['fuel'].loc[pathIndex]
                SpeedLegs = self.neighbour_graph['speed'].loc[pathIndex]

                SmoothedPath ={}
                SmoothedPath['type'] = 'Feature'
                SmoothedPath['geometry'] = {}
                SmoothedPath['geometry']['type'] = "LineString"
                SmoothedPath['geometry']['coordinates'] = nc.CrossingDF[['cx','cy']].to_numpy().tolist()            
                SmoothedPath['properties'] = {}
                SmoothedPath['properties']['from'] = Path['properties']['from']
                SmoothedPath['properties']['to']   = Path['properties']['to']
                SmoothedPath['properties']['traveltime'] = np.cumsum(TravelTimeLegs).tolist() 
                SmoothedPath['properties']['fuel']  = np.cumsum(FuelLegs).tolist()
                SmoothedPath['properties']['distance'] = np.cumsum(DistanceLegs).tolist()
                SmoothedPath['properties']['speed'] = SpeedLegs.tolist()
                SmoothedPaths.append(SmoothedPath)
                geojson['features'] = SmoothedPaths
                self.smoothed_paths = geojson
                self.mesh['paths'] = self.smoothed_paths
        for ii in range(len(self.mesh['paths']['features'])):
            self.mesh['paths']['features'][ii]['properties']['times'] = [str(ii) for ii in (pd.to_datetime(self.mesh['config']['Mesh_info']['Region']['startTime']) + pd.to_timedelta(self.mesh['paths']['features'][ii]['properties']['traveltime'],unit='days'))]

                    


'''
    Python scripts used to determine the optimal pathway of transients across the Mesh structure.
    This module can then be interchanged with alternative transient forecasts or the route paths from alternative
    ships/drones
'''


class Icebergs:
    """
        Simple module for processing and updating the iceberg locations through time determining the location on the
        same time interval asvthe PolarRoute Mesh, and returning the indices of the iceberg locations.
    """
    def __init__(self,cellGrid,object_locations):
        self.cellGrid         = cellGrid
        self.object_locations = object_locations

    def _cellGridResample(self):
        """
            Function that takes the super resolution of the iceberg tracking and resamples on the temporal sampling of
            the PolarRoute Mesh structure
        """
    
    def _physicsModel(self):
        """
            The physics based model describing the updating of the iceberg locations.
            After better methods are determined then this is the main area to update.

            INPUTS:


            OUTPUTS:
        """

    def forward(self):
        '''
            Given the Mesh and the iceberg locations forward propagate to determine the expected locations given a
            physics model describing the dynamics of the icebergs.
        '''

        






import argparse
import json
import inspect
import logging
from multiprocessing.connection import wait

from polar_route import __version__ as version
from polar_route.utils import setup_logging, timed_call


@setup_logging
def get_args(
        default_output: str,
        config_arg: bool = True,
        info_arg: bool = False,
        waypoints_arg: bool = False):
    """

    Parameters
    ----------
    config_arg
    info_arg

    Returns
    -------

    """
    ap = argparse.ArgumentParser()

    ap.add_argument("-o", "--output",
                    default=default_output,
                    help="Output file")
    ap.add_argument("-v", "--verbose",
                    default=False,
                    action="store_true",
                    help="Turn on DEBUG level logging")

    if config_arg:
        ap.add_argument("config", type=argparse.FileType("r"), 
                    help="File location of configuration file used to build the mesh")

    if info_arg:
        ap.add_argument("info", type=argparse.FileType("r"),
                    help="File location of the enviromental mesh")

    if waypoints_arg:
        ap.add_argument("waypoints", type=argparse.FileType("r"))
        ap.add_argument("-p", "--path_only",
                        default=False,
                        action = "store_true",
                        help="output only the calculated paths")

    return ap.parse_args()


@timed_call
def create_mesh_cli():
    """

    """
    from polar_route.mesh import Mesh

    args = get_args("create_mesh.output.json")
    logging.info("{} {}".format(inspect.stack()[0][3][:-4], version))

    config = json.load(args.config)

    # Discrete Meshing
    cg = Mesh(config)

    logging.info("Saving mesh to {}".format(args.output))
    info = cg.to_json()
    json.dump(info, open(args.output, "w"))


@timed_call
def add_vehicle_cli():
    """

    """
    from polar_route.vessel_performance import VesselPerformance

    args = get_args("add_vehicle.output.json", config_arg=False, info_arg=True)
    logging.info("{} {}".format(inspect.stack()[0][3][:-4], version))

    mesh = json.load(args.info)

    vp = VesselPerformance(mesh)

    logging.info("Saving mesh to {}".format(args.output))
    info = vp.to_json()
    json.dump(info, open(args.output, "w"))


@timed_call
def optimise_routes_cli():
    """

    """
    from polar_route.route_planner import RoutePlanner

    args = get_args("optimise_routes.output.json",
                    config_arg=False, info_arg=True, waypoints_arg= True)
    logging.info("{} {}".format(inspect.stack()[0][3][:-4], version))

    if args.path_only:
        logging.info("outputting only path to {}".format(args.output))
    else: 
        logging.info("outputting full mesh to {}".format(args.output))

    vehicle_mesh = json.load(args.info)
    rp = RoutePlanner(vehicle_mesh, args.waypoints)
    rp.compute_routes()
    rp.compute_smoothed_routes()
    info = rp.to_json()

    if args.path_only:
        json.dump(info['paths'], open(args.output, 'w'))
    else:
        json.dump(info, open(args.output, "w"))


__version__ = "0.0.13"
__description__ = "PolarRoute: Long-distance maritime polar route planning taking into account complex changing environmental conditions"
__license__ = "MIT"
__author__ = "Jonathan Smith, Samuel Hall, George Coombs, James Byrne,  Michael Thorne, Maria Fox"
__email__ = "jonsmi@bas.ac.uk"
__copyright__ = "2022, BAS AI Lab"



"""
The VesselPerformance class deals with all the vehicle specific features of the meshed environment model. It uses the
specific vessel parameters defined in the config to determine which cells in the mesh are inaccessible to a given
vehicle, either based on elevation or sea ice concentration, and determines vessel performance characteristics, such as
the speed and fuel consumption, based on the content of the cellboxes.

The input to the class is the mesh object produced by the CellGrid class in json format and it returns a modified mesh
in the same format via the VesselPerformance.to_json method.

Example:
    An example of how to use this code can be executed by running the following::

        import json
        from RoutePlanner.CellGrid import CellGrid
        from RoutePlanner.vessel_performance import VesselPerformance

        with open("config.json", 'r') as f:
            cfg = json.load(f)

        mesh = CellGrid(cfg)
        mesh_json = mesh.to_json()

        vp = VesselPerformance(mesh_json)

        vehicle_mesh = vp.to_json()
"""
import logging
import json
import numpy as np
import pandas as pd

from polar_route.utils import timed_call


class VesselPerformance:
    """
        Class for modelling the vessel performance.
        Takes a mesh as input in json format and modifies it to include vessel specifics.

        Attributes:
            mesh (dict): A dictionary containing all the mesh information
            config (dict): The config used to generate the input mesh
            mesh_df (DataFrame): The cellbox information from the mesh stored in a pandas DataFrame
            vessel_params (dict): The vessel specific information contained within the config

    """
    def __init__(self, mesh_json):
        """
            Constructs the VesselPerformance class from a given input mesh in json format which is then modified
            according to the vessel parameters defined in the config.

            Args:
                mesh_json (dict): The input mesh containing the cellboxes and neighbour graph as well as the config used
                to generate the mesh.
        """
        logging.info("Initialising vessel performance...")
        self.mesh = mesh_json
        self.config = self.mesh['config']
        self.mesh_df = pd.DataFrame(self.mesh['cellboxes']).set_index('id')
        self.vessel_params = self.config['Vessel']

        # Identifying land and extreme ice cells then removing them from the neighbour graph
        self.land()
        self.extreme_ice()
        self.inaccessible_nodes()

        # Checking if the speed is defined in the input mesh
        if 'speed' not in self.mesh_df:
            logging.debug(f'No speed in mesh, assigning default value of {self.vessel_params["Speed"]} '
                          f'{self.vessel_params["Unit"]} from config')
            self.mesh_df['speed'] = self.vessel_params["Speed"]

        # Modify speed based on ice resistance
        self.speed()

        # Calculate fuel usage based on speed and ice resistance
        self.fuel()

        # Check for NaNs and zero them then warn if present
        if self.mesh_df.isnull().values.any():
            logging.warning("NaNs present in mesh, setting all NaN values to zero!")
            self.mesh_df = self.mesh_df.fillna(0.)

        # Updating the mesh indexing and cellboxes
        self.mesh_df['id'] = self.mesh_df.index
        self.mesh['cellboxes'] = self.mesh_df.to_dict('records')

    def to_json(self):
        """
            Method to return the modified mesh in json format.

            Returns:
                j_mesh (dict): a dictionary representation of the modified mesh.
        """
        j_mesh = json.loads(json.dumps(self.mesh))
        return j_mesh

    def land(self):
        """
            Method to determine which cells are land based on configured minimum depth.
        """
        if 'elevation' not in self.mesh_df:
            logging.warning("No elevation data in mesh, no cells will be marked as land!")
            self.mesh_df['land'] = False
        else:
            self.mesh_df['land'] = self.mesh_df['elevation'] > self.vessel_params['MinDepth']
            logging.debug(f"{self.mesh_df['land'].sum()} cells inaccessible due to land")

    def extreme_ice(self):
        """
            Method to determine which cells are inaccessible based on configured max ice concentration.
        """
        if 'SIC' not in self.mesh_df:
            logging.info("No sea ice concentration data in mesh")
            self.mesh_df['ext_ice'] = False
        else:
            self.mesh_df['ext_ice'] = self.mesh_df['SIC'] > self.vessel_params['MaxIceExtent']
            logging.debug(f"{self.mesh_df['ext_ice'].sum()} cells inaccessible due to extreme ice")

    @timed_call
    def inaccessible_nodes(self):
        """
            Method to determine which nodes are inaccessible and remove them from the neighbour graph.
        """
        logging.info("Determining which nodes are inaccessible due to land and ice")
        inaccessible = self.mesh_df[(self.mesh_df['land']) | (self.mesh_df['ext_ice'])]
        inaccessible_idx = list(inaccessible.index)

        self.mesh['neighbour_graph'] = self.remove_nodes(self.mesh['neighbour_graph'], inaccessible_idx)

    def ice_resistance(self, velocity, area, thickness, density):
        """
            Method to find the ice resistance force at a given speed in a given cell.

            Args:
                velocity (float): The speed of the vessel in km/h
                area (float): The average sea ice concentration in the cell as a percentage
                thickness (float): The average ice thickness in the cell in m
                density (float): The average ice density in the cell in kg/m^3

            Returns:
                resistance (float): Resistance force in N
        """
        # Model parameters for different hull types
        hull_params = {'slender': [4.4, -0.8267, 2.0], 'blunt': [16.1, -1.7937, 3]}

        hull = self.vessel_params['HullType']
        beam = self.vessel_params['Beam']
        kparam, bparam, nparam = hull_params[hull]
        gravity = 9.81  # m/s-2

        speed = velocity * (5. / 18.)  # assume km/h and convert to m/s

        froude = speed / np.sqrt(gravity * area / 100 * thickness)
        resistance = 0.5 * kparam * (froude ** bparam) * density * beam * thickness * (speed ** 2) * (
                    (area / 100) ** nparam)
        return resistance

    def inverse_resistance(self, area, thickness, density):
        """
            Method to find the vessel speed that keeps the ice resistance force below a given threshold in a given cell.

            Args:
                area (float): The average sea ice concentration in the cell as a percentage
                thickness (float): The average ice thickness in the cell in m
                density (float): The average ice density in the cell in kg/m^3

            Returns:
                speed (float): Safe vessel speed in km/h
        """
        # Model parameters for different hull types
        hull_params = {'slender': [4.4, -0.8267, 2.0], 'blunt': [16.1, -1.7937, 3]}
        # force_limit (float): Resistance force value that should not be exceeded in N
        force_limit = self.vessel_params['ForceLimit']

        hull = self.vessel_params['HullType']
        beam = self.vessel_params['Beam']
        kparam, bparam, nparam = hull_params[hull]
        gravity = 9.81  # m/s-2

        vexp = 2 * force_limit / (kparam * density * beam * thickness * ((area / 100) ** nparam) * (
                    gravity * thickness * area / 100) ** -(bparam / 2))

        vms = vexp ** (1 / (2.0 + bparam))
        speed = vms * (18. / 5.)  # convert from m/s to km/h

        return speed

    @timed_call
    def speed(self):
        """
            Method to compile the new speeds calculated based on the ice resistance into the mesh.
        """
        logging.info("Calculating new speeds based on resistance models")

        if all(k in self.mesh_df for k in ("SIC", "thickness", "density")):
            logging.info("Adjusting speed according to ice resistance model")
            self.mesh_df['ice resistance'] = np.nan
            for idx, row in self.mesh_df.iterrows():
                if row['SIC'] == 0.0:
                    self.mesh_df.loc[idx, 'speed'] = self.vessel_params['Speed']
                    self.mesh_df.loc[idx, 'ice resistance'] = 0.0
                elif row['SIC'] > self.vessel_params['MaxIceExtent']:
                    self.mesh_df.loc[idx, 'speed'] = 0.0
                    self.mesh_df.loc[idx, 'ice resistance'] = np.inf
                else:
                    rp = self.ice_resistance(self.vessel_params['Speed'], row['SIC'], row['thickness'], row['density'])
                    if rp > self.vessel_params['ForceLimit']:
                        new_speed = self.inverse_resistance(row['SIC'], row['thickness'], row['density'])
                        rp = self.ice_resistance(new_speed, row['SIC'], row['thickness'], row['density'])
                        self.mesh_df.loc[idx, 'speed'] = new_speed
                        self.mesh_df.loc[idx, 'ice resistance'] = rp
                    else:
                        self.mesh_df.loc[idx, 'speed'] = self.vessel_params['Speed']
                        self.mesh_df.loc[idx, 'ice resistance'] = rp
        else:
            logging.info("No resistance data available, no speed adjustment necessary")

    def speed_simple(self):
        """
            Method to calculate the speed based on the sea ice concentration using a simple toy model.
        """
        self.mesh_df['speed'] = (1 - np.sqrt(self.mesh_df['SIC'] / 100)) * \
                                        self.vessel_params['Speed']

    @timed_call
    def fuel(self):
        """
            Method to calculate the fuel usage in tons per day based on speed in km/h and ice resistance in N.
        """
        logging.info("Calculating fuel requirements in each cell")

        if 'ice resistance' in self.mesh_df:
            self.mesh_df['fuel'] = (0.00137247 * self.mesh_df['speed'] ** 2 - 0.0029601 *
                                    self.mesh_df['speed'] + 0.25290433
                                    + 7.75218178e-11 * self.mesh_df['ice resistance'] ** 2
                                    + 6.48113363e-06 * self.mesh_df['ice resistance']) * 24.0
        else:
            self.mesh_df['fuel'] = (0.00137247 * self.mesh_df['speed'] ** 2 - 0.0029601 *
                                    self.mesh_df['speed'] + 0.25290433) * 24.0

    def remove_nodes(self, neighbour_graph, inaccessible_nodes):
        """
            Method to remove a list of inaccessible nodes from a given neighbour graph.

            Args:
                neighbour_graph (dict): A dictionary containing indexes of cellboxes and how they are connected

                {
                    'index':{
                        '1':[index,...],
                        '2':[index,...],
                        '3':[index,...],
                        '4':[index,...],
                        '-1':[index,...],
                        '-2':[index,...],
                        '-3':[index,...],
                        '-4':[index,...]
                    },
                    'index':{...},
                    ...
                }

                inaccessible_nodes (list): A list of indexes to be removed from the neighbour_graph

            Returns:
                accessibility_graph (dict): A new neighbour graph with the inaccessible nodes removed
        """
        logging.debug(f"Removing {len(inaccessible_nodes)} nodes from the neighbour graph")
        accessibility_graph = neighbour_graph.copy()

        for node in accessibility_graph.keys():
            for case in accessibility_graph[node].keys():
                for inaccessible_node in inaccessible_nodes:
                    if int(inaccessible_node) in accessibility_graph[node][case]:
                        accessibility_graph[node][case].remove(int(inaccessible_node))

        for node in inaccessible_nodes:
            accessibility_graph.pop(node)

        return accessibility_graph



"""
    The python package `crossing` implement the optimisation for the crossing point for the unsmoothed and smoothed path
     construction. The package is separated into two classes `NewtonianDistance` and `NewtonianCurve`.
     In the section below we will go through, stage by stage, how the crossing point is determined and the methods
     used within the classes.

"""
import copy
import pandas as pd
import numpy as np
np.seterr(divide='ignore', invalid='ignore')


class NewtonianDistance:
    """The summary line for a class docstring should fit on one line.

    If the class has public attributes, they may be documented here
    in an ``Attributes`` section and follow the same formatting as a
    function's ``Args`` section. Alternatively, attributes may be documented
    inline with the attribute's declaration (see __init__ method below).

    Properties created with the ``@property`` decorator should be documented
    in the property's getter method.

    Attributes:
        attr1 (str): Description of `attr1`.
        attr2 (:obj:`int`, optional): Description of `attr2`.

    """
    def __init__(self,source_graph=None,neighbour_graph=None,
                 case=None,unit_shipspeed='km/hr',unit_time='days',
                 zerocurrents=True,debugging=False,maxiter=1000,optimizer_tol=1e-3):
        """Example of docstring on the __init__ method.

        The __init__ method may be documented in either the class level
        docstring, or as a docstring on the __init__ method itself.

        Either form is acceptable, but the two should not be mixed. Choose one
        convention to document the __init__ method and be consistent with it.

        Note:
            Do not include the `self` parameter in the ``Args`` section.

        Args:
            param1 (str): Description of `param1`.
            param2 (:obj:`int`, optional): Description of `param2`. Multiple
                lines are supported.
            param3 (:obj:`list` of :obj:`str`): Description of `param3`.

        """
        # Cell information
        self.source_graph     = source_graph
        self.neighbour_graph  = neighbour_graph

        # Inside the code the base units are m/s. Changing the units of the inputs to match
        self.unit_shipspeed  = unit_shipspeed
        self.unit_time       = unit_time
        self.source_speed    = self._unit_speed(self.source_graph['Speed'])
        self.neighbour_speed = self._unit_speed(self.neighbour_graph['Speed'])
        self.case            = case

        if zerocurrents:
            self.zero_current_factor = 0.0
        else:
            self.zero_current_factor = 1.0

        # Optimisation Information
        self.maxiter       = maxiter
        self.optimizer_tol = optimizer_tol

        # Optimisation Information
        self.m_long  = 111.321*1000
        self.m_lat   = 111.386*1000

        # Defining a small distance
        self.small_distance = 1e-4

        # For Debugging purposes
        self.debugging     = debugging

    def _newton_optimisation(self, f, x, a, Y, u1, v1, u2, v2, s1, s2):
        '''
            FILL
        '''
        y0 = (Y*x)/(x+a)
        if self.debugging:
            print('---Initial y={:.2f}'.format(y0))
        improving = True
        iterartion_num = 0
        while improving:
            F,dF,X1,X2,t1,t2  = f(y0,x,a,Y,u1,v1,u2,v2,s1,s2)
            if self.debugging:
                print('---Iteration {}: y={:.2f}; F={:.5f}; dF={:.2f}'\
                      .format(iterartion_num,y0,F,dF))
            y0  = y0 - (F/dF)
            improving = abs((F/dF)/(X1*X2)) > self.optimizer_tol
            iterartion_num+=1
            if iterartion_num>1000:
                raise Exception('Newton not able to converge')
        return y0,self._unit_time(np.array([t1,t2]))

    def _unit_speed(self,val):
        '''
            FILL
        '''
        if not isinstance(val,type(None)):
            if self.unit_shipspeed == 'km/hr':
                val = val*(1000/(60*60))
            if self.unit_shipspeed == 'knots':
                val = (val*0.51)
            return val
        else:
            return None

    def _unit_time(self,val):
        '''
            FILL
        '''
        if self.unit_time == 'days':
            val = val/(60*60*24.)
        elif self.unit_time == 'hr':
            val = val/(60*60.)
        elif self.unit_time == 'min':
            val = val/(60.)
        elif self.unit_time == 's':
            val = val
        return val

    def waypoint_correction(self,Wp,Cp):
        '''
            FILL
        '''
        x = (Cp[0]-Wp[0])*self.m_long*np.cos(Wp[1]*(np.pi/180))
        y = (Cp[1]-Wp[1])*self.m_lat
        Su  = self.source_graph['Vector_x']*self.zero_current_factor
        Sv  = self.source_graph['Vector_y']*self.zero_current_factor
        Ssp = self.source_speed
        traveltime = self._traveltime_in_cell(x,y,Su,Sv,Ssp)
        return self._unit_time(traveltime)

    def _F(self,y,x,a,Y,u1,v1,u2,v2,s1,s2):
        '''
            FILL
        '''
        C1 = s1**2 - u1**2 - v1**2
        C2 = s2**2 - u2**2 - v2**2
        D1 = x*u1 + y*v1
        D2 = a*u2 + (Y-y)*v2
        X1ns = D1**2 + C1*(x**2 + y**2)
        if X1ns < 0:
            X1 = -1
        else:
            X1 = np.sqrt(X1ns)
        X2ns = D2**2 + C2*(a**2 + (Y-y)**2)
        if X2ns < 0:
            X2 = -1
        else:
            X2 = np.sqrt(X2ns)

        F  = X2*(y-((v1*(X1-D1))/C1)) + X1*(y-Y+((v2*(X2-D2))/C2))

        dD1 = v1
        dD2 = -v2
        if X1 == 0:
            dX1 = 0
        else:
            dX1 = (D1*v1 + C1*y)/X1
        if X2 ==0:
            dX2 = 0
        else:
            dX2 = (-D2*v2 - C2*(Y-y))/X2
        dF  = (X1+X2) + y*(dX1 + dX2) - (v1/C1)*(dX2*(X1-D1) + X2*(dX1-dD1))\
            + (v2/C2)*(dX1*(X2-D2)+X1*(dX2-dD2)) - Y*dX1

        t1 = (X1-D1)/C1
        t2 = (X2-D2)/C2

        return F,dF,X1,X2,t1,t2

    def _traveltime_in_cell(self,xdist,ydist,U,V,S):
        '''
            Determines the travel-time within cell between two points
        '''
        dist  = np.sqrt(xdist**2 + ydist**2)
        cval  = np.sqrt(U**2 + V**2)

        dotprod  = xdist*U + ydist*V
        diffsqrs = S**2 - cval**2

        # if (dotprod**2 + diffsqrs*(dist**2) < 0)
        if diffsqrs == 0.0:
            if dotprod == 0.0:
                return np.inf
                #raise Exception(' ')
            else:
                if ((dist**2)/(2*dotprod))  <0:
                    return np.inf
                    #raise Exception(' ')
                else:
                    traveltime = dist * dist / (2 * dotprod)
                    return traveltime

        traveltime = (np.sqrt(dotprod**2 + (dist**2)*diffsqrs) - dotprod)/diffsqrs
        if traveltime < 0:
            traveltime = np.inf
            #raise Exception('Newton Corner Cases returning Zero Traveltime - ISSUE')
        return traveltime

    def _longitude(self):
        '''
            FILL
        '''

        if self.case==2:
            ptvl = 1.0
        else:
            ptvl = -1.0

        s_cx  = self.source_graph['cx']
        s_cy  = self.source_graph['cy']
        s_dcx = self.source_graph['dcx']
        s_dcy = self.source_graph['dcy']
        n_cx  = self.neighbour_graph['cx']
        n_cy  = self.neighbour_graph['cy']
        n_dcx = self.neighbour_graph['dcx']
        n_dcy = self.neighbour_graph['dcy']


        Su = ptvl*self.source_graph['Vector_x']*self.zero_current_factor
        Sv = ptvl*self.source_graph['Vector_y']*self.zero_current_factor
        Nu = ptvl*self.neighbour_graph['Vector_x']*self.zero_current_factor
        Nv = ptvl*self.neighbour_graph['Vector_y']*self.zero_current_factor

        Ssp = self.source_speed
        Nsp = self.neighbour_speed


        x = s_dcx*self.m_long*np.cos(s_cy*(np.pi/180))
        a = n_dcx*self.m_long*np.cos(n_cy*(np.pi/180))
        Y = ptvl*(n_cy-s_cy)*self.m_lat

        # Optimising to determine the y-value of the crossing point
        y,TravelTime = self._newton_optimisation(self._F,x,a,Y,Su,Sv,Nu,Nv,Ssp,Nsp)
        CrossPoints = (s_cx+ptvl*s_dcx,\
                       s_cy+ptvl*y/self.m_lat)
        CellPoints  = [n_cx,n_cy]

        return TravelTime,CrossPoints,CellPoints

    def _latitude(self):
        '''
            FILL
        '''

        if self.case==4:
            ptvl = 1.0
        else:
            ptvl = -1.0

        s_cx  = self.source_graph['cx']
        s_cy  = self.source_graph['cy']
        s_dcx = self.source_graph['dcx']
        s_dcy = self.source_graph['dcy']
        n_cx  = self.neighbour_graph['cx']
        n_cy  = self.neighbour_graph['cy']
        n_dcx = self.neighbour_graph['dcx']
        n_dcy = self.neighbour_graph['dcy']


        Su = -1*ptvl*self.source_graph['Vector_y']*self.zero_current_factor
        Sv = ptvl*self.source_graph['Vector_x']*self.zero_current_factor
        Nu = -1*ptvl*self.neighbour_graph['Vector_y']*self.zero_current_factor
        Nv = ptvl*self.neighbour_graph['Vector_x']*self.zero_current_factor

        Ssp=self.source_speed
        Nsp=self.neighbour_speed

        x = s_dcy*self.m_lat
        a = n_dcy*self.m_lat
        Y = ptvl*(n_cx-s_cx)*self.m_long*np.cos((n_cy+s_cy)*(np.pi/180)/2.0)

        y,TravelTime   = self._newton_optimisation(self._F,x,a,Y,Su,Sv,Nu,Nv,Ssp,Nsp)
        clon = s_cx  + ptvl*y/(self.m_long*np.cos((n_cy+\
               s_cy)*(np.pi/180)/2.0))
        clat = s_cy + -1*ptvl*s_dcy

        CrossPoints = (clon,clat)
        CellPoints  = [n_cx,n_cy]

        return TravelTime,CrossPoints,CellPoints


    def _corner(self):
        '''
            FILL
        '''


        s_cx  = self.source_graph['cx']
        s_cy  = self.source_graph['cy']
        s_dcx = self.source_graph['dcx']
        s_dcy = self.source_graph['dcy']
        n_cx  = self.neighbour_graph['cx']
        n_cy  = self.neighbour_graph['cy']
        n_dcx = self.neighbour_graph['dcx']
        n_dcy = self.neighbour_graph['dcy']



        # Given the determine the postive and negative position relative to centre
        if self.case==1:
            ptvX = 1.0
            ptvY = 1.0
        elif self.case==-1:
            ptvX = -1.0
            ptvY = -1.0
        elif self.case==3:
            ptvX = 1.0
            ptvY = -1.0
        elif self.case==-3:
            ptvX = -1.0
            ptvY = 1.0

        dx1 = s_dcx*self.m_long*np.cos(s_cy*(np.pi/180))
        dx2 = n_dcx*self.m_long*np.cos(n_cy*(np.pi/180))
        dy1 = s_dcy*self.m_lat
        dy2 = n_dcy*self.m_lat

        # Currents in Cells
        Su = ptvX*self.source_graph['Vector_x']*self.zero_current_factor
        Sv = ptvY*self.source_graph['Vector_y']*self.zero_current_factor
        Nu = ptvX*self.neighbour_graph['Vector_x']*self.zero_current_factor
        Nv = ptvY*self.neighbour_graph['Vector_y']*self.zero_current_factor

        # Vehicles Speeds in Cells
        Ssp = self.source_speed; Nsp = self.neighbour_speed

        # Determining the crossing point as the corner of the case
        CrossPoints = [s_cx+ptvX*s_dcx,\
                       s_cy+ptvY*s_dcy]
        CellPoints  = [n_cx,n_cy]

        # Determining traveltime
        t1 = self._traveltime_in_cell(dx1,dy1,Su,Sv,Ssp)
        t2 = self._traveltime_in_cell(dx2,dy2,Nu,Nv,Nsp)
        TravelTime  = self._unit_time(np.array([t1,t2]))

        return TravelTime,CrossPoints,CellPoints


    def value(self):
        '''
            FILLE
        '''
        if self.debugging:
            print('============================================')
        if abs(self.case)==2:
            TravelTime,CrossPoints,CellPoints = self._longitude()
        elif abs(self.case)==4:
            TravelTime,CrossPoints,CellPoints = self._latitude()
        elif abs(self.case)==1 or abs(self.case)==3:
            TravelTime,CrossPoints,CellPoints = self._corner()
        else:
            print('---> Issue with cell (Xsc,Ysc)={:.2f};{:.2f}'.\
                format(self.source_graph['cx'],self.source_graph['cy']))
            TravelTime  = [np.inf,np.inf]
            CrossPoints = [np.nan,np.nan]
            CellPoints  = [np.nan,np.nan]

        return TravelTime, CrossPoints, CellPoints


# ===================================================================================================
# ===================================================================================================
# ===================================================================================================
# ===================================================================================================
# ===================================================================================================
# ===================================================================================================


class NewtonianCurve:
    def __init__(self,neighbour_graph,config,unit_shipspeed='km/hr',unit_time='days',debugging=False,maxiter=1000,pathIter=5,optimizer_tol=1e-3,minimumDiff=1e-3,zerocurrents=True):
        '''
            FILL
        '''
        # Passing the Dijkstra Graph
        self.neighbour_graph = copy.copy(neighbour_graph)

        # Passing the optional Information
        self.config = config

        
        # Inside the code the base units are m/s. Changing the units of the inputs to match
        self.unit_shipspeed = unit_shipspeed
        self.unit_time      = unit_time

        
        # Information for distance metrics
        self.R              = 6371.*1000

        # Optimisation Information
        self.maxiter       = maxiter
        self.pathIter      = pathIter
        self.optimizer_tol = optimizer_tol
        self.minimumDiff   = minimumDiff
        self._epsilon      = 1e-3


        # Optimisation Information
        self.m_long  = 111.321*1000
        self.m_lat   = 111.386*1000.

        # For Debugging purposes 
        self.debugging     = debugging

        if self.debugging:
            self.debugFile1 = open("debugFil.txt", "w")  # append mode
            self.debugFile1.write("Today \n")

        self.id = 0

        # zeroing currents if flagged
        if zerocurrents:
            self.zc = 0.0
        else:
            self.zc = 1.0

    def _unit_speed(self,Val):
        '''
            FILL
        '''
        if self.unit_shipspeed == 'km/hr':
            Val = Val*(1000/(60*60))
        if self.unit_shipspeed == 'knots':
            Val = (Val*0.51)
        return Val

    def _unit_time(self,Val):
        '''
            FILL
        '''
        if self.unit_time == 'days':
            Val = Val/(60*60*24)
        elif self.unit_time == 'hr':
            Val = Val/(60*60)
        elif self.unit_time == 'min':
            Val = Val/(60)
        elif self.unit_time == 's':
            Val = Val
        return Val


    def _calXDist(self,start_long,end_long):
        '''
            FILL
        '''
        return (end_long - start_long)*self.m_long#*np.cos(centralLat)
    def _calYDist(self,start_lat,end_lat):
        '''
            FILL
        '''
        return (end_lat-start_lat)*self.m_lat

    def _long_case(self):
        '''
            FILL
        '''
        def NewtonOptimisationLong(f,y0,x,a,Y,u1,v1,u2,v2,speed_s,speed_e,R,λ_s,φ_r):
                tryNum=1
                iter=0
                improving=True
                while improving:  
                    F,dF,X1,X2  = f(y0,x,a,Y,u1,v1,u2,v2,speed_s,speed_e,R,λ_s,φ_r)
                    if (F==0) or (dF==0):
                        dY = 0
                    else:
                        dY = (F/dF)
                    improving =  (abs(dY)>self._epsilon) or (abs(dY) > self._epsilon*(X1*X2) and (abs(dY)/iter) > self._epsilon)
                    y0  -= dY
                    iter+=1

                    if (iter>100 and tryNum == 1):
                        y0 = Y*x/(x+a)
                        tryNum+=1
                    if (iter > 200) and tryNum>= 2 and tryNum < 10:
                        tryNum+=1
                        iter-=100
                        if(Y < 0):
                            if v2>v1:
                                y0 = (tryNum-2)*Y
                            else:
                                y0 = (tryNum-3)*-Y
                        else:
                            if (v2<v1):
                                y0 = (tryNum-2)*Y
                            else:
                                y0 = (tryNum-3)*-Y
                    if iter > 1000:
                        raise Exception('Newton Curve Issue')
                return y0

        def _F(y,x,a,Y,u1,v1,u2,v2,speed_s,speed_e,R,λ_s,φ_r):
            θ  = (y/R + λ_s*(np.pi/180))
            zl = x*np.cos(θ)
            ψ  = (-(Y-y)/R + φ_r*(np.pi/180))
            zr = a*np.cos(ψ)

            C1  = speed_s**2 - u1**2 - v1**2
            C2  = speed_e**2 - u2**2 - v2**2
            D1  = zl*u1 + y*v1
            D2  = zr*u2 + (Y-y)*v2
            X1  = np.sqrt(D1**2 + C1*(zl**2 + y**2))
            X2  = np.sqrt(D2**2 + C2*(zr**2 + (Y-y)**2))

            dzr = -zr*np.sin(ψ)/R
            dzl = -zl*np.sin(θ)/R

            dD1 = dzl*u1 + v1
            dD2 = dzr*u2 - v2
            dX1 = (D1*v1 + C1*y + dzl*(D1*u1 + C1*zl))/X1
            dX2 = (-v2*D2 - C2*(Y-y) + dzr*(D2*u2 + C2*zr))/X2     

            zr_term = (zr - (X2 - D2)*u2/C2)
            zl_term = (zl - (X1 - D1)*u1/C1)

            F = (X1+X2)*y - v1*(X1-D1)*X2/C1 - (Y - v2*(X2-D2)/C2)*X1 + dzr*zr_term*X1 + dzl*zl_term*X2

            dF = (X1+X2) + y*(dX1 + dX2) - (v1/C1)*(dX2*(X1-D1) + X2*(dX1-dD1))\
                - Y*dX1 + (v2/C2)*(dX1*(X2-D2) + X1*(dX2-dD2))\
                - (zr/(R**2))*zr_term*X1\
                - (zl/(R**2))*zl_term*X2\
                + dzr*(dzr-u2*(dX2-dD2))/C2*X1\
                + dzl*(dzl-u1*(dX1-dD1))/C1*X2\
                + dzr*zr_term*dX1 + dzl*zl_term*dX2

            return F,dF,X1,X2

        Sp   = tuple(self.triplet[['cx','cy']].iloc[0])
        Cp   = tuple(self.triplet[['cx','cy']].iloc[1])
        Np   = tuple(self.triplet[['cx','cy']].iloc[2])

        cell_s_u = self.neighbour_graph.loc[self.triplet.iloc[1]['cellStart'].name,'Vector_x']
        cell_s_v = self.neighbour_graph.loc[self.triplet.iloc[1]['cellStart'].name,'Vector_y']
        cell_e_u = self.neighbour_graph.loc[self.triplet.iloc[1]['cellEnd'].name,'Vector_x']
        cell_e_v = self.neighbour_graph.loc[self.triplet.iloc[1]['cellEnd'].name,'Vector_y']

        if self.triplet.iloc[1].case == 2:   
            sgn  = 1
        else:
            sgn  = -1

        λ_s  = Sp[1]
        φ_r  = Np[1]

        x           = sgn*self._calXDist(Sp[0],Cp[0])
        a           = sgn*self._calXDist(Cp[0],Np[0])
        Y           = (Np[1]-Sp[1])*self.m_lat
        y0          = Y/2
        u1          = sgn*self.zc*cell_s_u; v1 = self.zc*cell_s_v
        u2          = sgn*self.zc*cell_e_u; v2 = self.zc*cell_e_v
        y           = NewtonOptimisationLong(_F,y0,x,a,Y,u1,v1,u2,v2,self.speed_s,self.speed_e,self.R,λ_s,φ_r)

        # Updating the crossing points
        self.triplet['cx'].iloc[1] = Cp[0]
        self.triplet['cy'].iloc[1] = Sp[1] + y/self.m_lat


    def _lat_case(self):
        '''
            FILL
        '''
        def NewtonOptimisationLat(f,y0,x,a,Y,u1,v1,u2,v2,speed_s,speed_e,R,λ,θ,ψ):
                tryNum=1
                iter=0
                improving=True
                while improving:  
                    F,dF,X1,X2  = f(y0,x,a,Y,u1,v1,u2,v2,speed_s,speed_e,R,λ,θ,ψ)
                    if (F==0) or (dF==0):
                        dY = 0
                    else:
                        dY = (F/dF)
                    improving =abs(dY) > 1 or (abs(dY) > self._epsilon*(X1*X2) and (abs(dY)/iter) > self._epsilon)
                    y0  -= dY
                    iter+=1

                    if (iter>100 and tryNum == 1):
                        y0 = Y*x/(x+a)
                        tryNum+=1
                    if (iter > 200) and tryNum== 2:
                        tryNum+=1
                        if(Y < 0):
                            if v2>v1:
                                y0 = Y
                            else:
                                y0 = 0
                        else:
                            if (v2<v1):
                                y0 = Y
                            else:
                                y0 = 0
                    if iter > 1000:
                        raise Exception('Newton Curve Issue')
                return y0

        def _F(y,x,a,Y,u1,v1,u2,v2,speed_s,speed_e,R,λ,θ,ψ):
            λ   = λ*(np.pi/180)
            ψ   = ψ*(np.pi/180)
            θ   = θ*(np.pi/180)
            r1  = np.cos(λ)/np.cos(θ)
            r2  = np.cos(ψ)/np.cos(θ)

            d1  = np.sqrt(x**2 + (r1*y)**2)
            d2  = np.sqrt(a**2 + (r2*(Y-y))**2)
            C1  = speed_s**2 - u1**2 - v1**2
            C2  = speed_e**2 - u2**2 - v2**2
            D1  = x*u1 + r1*v1*y
            D2  = a*u2 + r2*v2*(Y-y)
            X1  = np.sqrt(D1**2 + C1*(d1**2))
            X2  = np.sqrt(D2**2 + C2*(d2**2)) 

            dX1 = (r1*(D1*v1 + r1*C1*y))/X1

            # Cassing Warning message - FIX !!
            dX2 = (-r2*(D2*v2 + r2*C2*(Y-y)))/X2

            F = ((r2**2)*X1+(r1**2)*X2)*y - r1*v1*(X1-D1)*X2/C1 - r2*(r2*Y-v2*(X2-D2)/C2)*X1

            dF = ((r2**2)*X1 + (r1**2)*X2) + y*((r2**2)*dX1 + (r1**2)*dX2)\
                - r1*v1*((X1-D1)*dX2 + (dX1-r1*v1)*X2)/C1\
                - (r2**2)*Y*dX1\
                + r2*v2*((X2-D2)*dX1 + (dX2+r2*v2)*X1)/C2

            return F,dF,X1,X2


        Sp = tuple(self.triplet.iloc[0][['cx','cy']])
        Cp = tuple(self.triplet.iloc[1][['cx','cy']])
        Np = tuple(self.triplet.iloc[2][['cx','cy']])

        cell_s_u = self.neighbour_graph.loc[self.triplet.iloc[1]['cellStart'].name,'Vector_x']
        cell_s_v = self.neighbour_graph.loc[self.triplet.iloc[1]['cellStart'].name,'Vector_y']
        cell_e_u = self.neighbour_graph.loc[self.triplet.iloc[1]['cellEnd'].name,'Vector_x']
        cell_e_v = self.neighbour_graph.loc[self.triplet.iloc[1]['cellEnd'].name,'Vector_y']

        if self.triplet.iloc[1].case == 4:   
            sgn   = 1
        else:
            sgn   = -1

        λ=Sp[1]
        θ=Cp[1]   
        ψ=Np[1]  

        x     = sgn*self._calYDist(Sp[1],Cp[1])
        a     = sgn*self._calYDist(Cp[1],Np[1])
        Y     = sgn*(Np[0]-Sp[0])*self.m_long*np.cos(Cp[1]*(np.pi/180))
        Su    = -sgn*self.zc*cell_s_v; Sv = sgn*self.zc*cell_s_u
        Nu    = -sgn*self.zc*cell_e_v; Nv = sgn*self.zc*cell_e_u
        y0    = Y/2

        y     = NewtonOptimisationLat(_F,y0,x,a,Y,Su,Sv,Nu,Nv,self.speed_s,self.speed_e,self.R,λ,θ,ψ)

        self.triplet['cx'].iloc[1] = Sp[0] + sgn*y/(self.m_long*np.cos(Cp[1]*(np.pi/180)))
        self.triplet['cy'].iloc[1] = Cp[1]


    def _corner_case(self):
        '''
            FILL
        '''
        # Separting out the Long/Lat of each of the points
        Xs,Ys = tuple(self.triplet.iloc[0][['cx','cy']])
        Xc,Yc = tuple(self.triplet.iloc[1][['cx','cy']])
        Xe,Ye = tuple(self.triplet.iloc[2][['cx','cy']])

        # === 1. Assess the cells that are shared commonly in the corner case ====
        sourceNeighbourIndices = self.triplet.iloc[1]['cellStart']
        endNeighbourIndices    = self.triplet.iloc[1]['cellEnd']
    
        commonIndices = list(set(sourceNeighbourIndices['neighbourIndex']).intersection(endNeighbourIndices['neighbourIndex']))
        CornerCells   = self.neighbour_graph.loc[commonIndices]
        Y_line = ((Ye-Ys)/(Xe-Xs))*(Xc-Xs) + Ys

        try:
            if Yc >= Y_line:
                newCell = CornerCells.loc[CornerCells['cy'].idxmin()]
                if newCell.cy > Yc:
                    return
            elif Yc < Y_line:
                newCell = CornerCells.loc[CornerCells['cy'].idxmax()]
                if newCell.cy < Yc:
                    return
            # === 3. Return the path crossing points and cell indices

            firstCrossingPoint  = np.array(sourceNeighbourIndices['neighbourCrossingPoints'])[np.where(np.array(sourceNeighbourIndices['neighbourIndex'])==newCell.name)[0][0],:]
            secondCrossingPoint = np.array(newCell['neighbourCrossingPoints'])[np.where(np.array(newCell['neighbourIndex'])==endNeighbourIndices.name)[0][0],:]
        except:
            self.triplet = copy.deepcopy(self.org_triplet)
            return

        # Adding in the new crossing Point
        newP = pd.Series(name=self.triplet.iloc[1].name+1,dtype='object')
        newP['cx']        = secondCrossingPoint[0]
        newP['cy']        = secondCrossingPoint[1]
        newP['cellStart'] = newCell
        newP['cellEnd']   = copy.deepcopy(self.triplet['cellEnd'].iloc[1])
        newP['case']      = newP['cellStart']['case'][np.where(np.array(newP['cellStart']['neighbourIndex'])==newP['cellEnd'].name)[0][0]]


        # Updating the origional crossing point
        self.triplet['cx'].iloc[1]      = firstCrossingPoint[0]
        self.triplet['cy'].iloc[1]      = firstCrossingPoint[1]
        self.triplet['cellEnd'].iloc[1] = newCell 
        self.triplet['case'].iloc[1]    = self.triplet['cellStart'].iloc[1]['case'][np.where(np.array(self.triplet['cellStart'].iloc[1]['neighbourIndex'])==newCell.name)[0][0]]


        # Adding the new crossing point to the triplet
        self.newP = newP

        self.CrossingDF = pd.concat([self.CrossingDF,newP.to_frame().transpose()]).sort_index().reset_index(drop=True) #self.CrossingDF.append(newP,sort=True).sort_index().reset_index(drop=True)
        self.CrossingDF.index = np.arange(int(self.CrossingDF.index.min()),int(self.CrossingDF.index.max()*1e3 + 1e3),int(1e3))


    def _mergePoint(self):
        '''
            FILL
        '''
        def PtDist(Ser1,Ser2):
            #return np.sqrt(self._calXDist(Ser2['cx'],Ser1['cx'])**2 + self._calXDist(Ser2['cy'],Ser1['cy'])**2)
            return np.sqrt((Ser1['cx'] - Ser2['cx'])**2 + (Ser1['cy'] - Ser2['cy'])**2)

        id=0
        while id < len(self.CrossingDF)-3:
            triplet = self.CrossingDF.iloc[id:id+3]
            if PtDist(triplet.iloc[0],triplet.iloc[1]) < 2e-3:
                try:
                    neighbourIndex = np.where(np.array(triplet.iloc[0]['cellStart']['neighbourIndex'])==triplet.iloc[1]['cellEnd'].name)[0][0]
                except:
                    id+=1
                    continue
                case           = triplet['cellStart'].iloc[0]['case'][neighbourIndex]
                crossingPoint  = triplet['cellStart'].iloc[0]['neighbourCrossingPoints'][neighbourIndex]
                triplet['cx'].iloc[0]      = crossingPoint[0]
                triplet['cy'].iloc[0]      = crossingPoint[1]
                triplet['cellEnd'].iloc[0] = copy.deepcopy(triplet.iloc[1]['cellEnd'])
                triplet['case'].iloc[0]    = copy.deepcopy(case)
                self.CrossingDF           = self.CrossingDF.drop(triplet.iloc[1].name)
            if PtDist(triplet.iloc[1],triplet.iloc[2]) < 2e-3:
                try:
                    neighbourIndex = np.where(np.array(triplet.iloc[1]['cellStart']['neighbourIndex'])==triplet.iloc[2]['cellEnd'].name)[0][0]
                except:
                    id+=1
                    continue
                case           = triplet['cellStart'].iloc[1]['case'][neighbourIndex]
                crossingPoint  = triplet['cellStart'].iloc[1]['neighbourCrossingPoints'][neighbourIndex]
                triplet['cx'].iloc[1]      = crossingPoint[0]
                triplet['cy'].iloc[1]      = crossingPoint[1]
                triplet['cellEnd'].iloc[1] = copy.deepcopy(triplet.iloc[2]['cellEnd'])
                triplet['case'].iloc[1]    = copy.deepcopy(case)
                self.CrossingDF           = self.CrossingDF.drop(triplet.iloc[2].name)

            id+=1

        self.CrossingDF = self.CrossingDF.sort_index().reset_index(drop=True)
        self.CrossingDF.index = np.arange(int(self.CrossingDF.index.min()),int(self.CrossingDF.index.max()*1e3 + 1e3),int(1e3))



    def _horseshoe(self):
        '''
            FILL
        '''
        # Defining the case information
        Cp             = tuple(self.triplet[['cx','cy']].iloc[1])
        Sp             = tuple(self.triplet.iloc[0][['cx','cy']])
        Np             = tuple(self.triplet.iloc[2][['cx','cy']])
        
        cellStartGraph = self.triplet.iloc[1]['cellStart']
        cellEndGraph   = self.triplet.iloc[1]['cellEnd']        
        case           = self.triplet['case'].iloc[1]

        # Returning if corner horseshoe case type
        if abs(case)==1 or abs(case)==3 or abs(case)==0: 
            return
        elif abs(case) == 2:
            # Defining the min and max of the start and end cells
            smin = cellStartGraph['cy']-cellStartGraph['dcy'] 
            smax = cellStartGraph['cy']+cellStartGraph['dcy']
            emin = cellEndGraph['cy']-cellEndGraph['dcy']
            emax = cellEndGraph['cy']+cellEndGraph['dcy']

            # Defining the global min and max
            vmin = np.max([smin,emin])
            vmax = np.min([smax,emax])

            # Point crossingpoint on boundary between the two origional cells
            if (Cp[1] > vmin) and (Cp[1] < vmax):
                return

            # If Start and end cells share a edge for the horesshoe 
            if (Cp[1]<=smin) and (smin==emin):
                hrshCaseStart = 4
                hrshCaseEnd   = 4
            if (Cp[1]>=smax) and (smax==emax):
                hrshCaseStart = -4
                hrshCaseEnd   = -4

            # --- Cases where StartCell is Larger than end Cell ---
            if (Cp[1]>=emax) and (smax>emax):
                hrshCaseStart = case
                hrshCaseEnd   = (-4)                
            if (Cp[1]<=emin) and (smin<emin):
                hrshCaseStart = case
                hrshCaseEnd   = (4)                   

            # --- Cases where StartCell is smaller than end Cell ---
            if (Cp[1]>=smax) and (smax<emax):
                hrshCaseStart = -4
                hrshCaseEnd   = -case
            if (Cp[1]<=smin) and (emin<smin):
                hrshCaseStart = 4
                hrshCaseEnd   = -case                    

        elif abs(case) == 4:
            # Defining the min and max of the start and end cells
            smin = cellStartGraph['cx']-cellStartGraph['dcx']
            smax = cellStartGraph['cx']+cellStartGraph['dcx']
            emin = cellEndGraph['cx']-cellEndGraph['dcx']
            emax = cellEndGraph['cx']+cellEndGraph['dcx']

            # Defining the global min and max
            vmin = np.max([smin,emin])
            vmax = np.min([smax,emax])

            # Point crossingpoint on boundary between the two origional cells
            if (Cp[0] >= vmin) and (Cp[0] <= vmax):
                return

            # If Start and end cells share a edge for the horesshoe 
            if (Cp[0]<smin) and (smin==emin):
                hrshCaseStart = -2
                hrshCaseEnd   = -2
            if (Cp[0]>smax) and (smax==emax):
                hrshCaseStart = 2
                hrshCaseEnd   = 2

            # --- Cases where StartCell is Larger than end Cell ---
            if (Cp[0]>emax) and (smax>emax):
                hrshCaseStart = case
                hrshCaseEnd   = (2)                
            if (Cp[0]<emin) and (smin<emin):
                hrshCaseStart = case
                hrshCaseEnd   = (-2)                   

            # --- Cases where StartCell is smaller than end Cell ---
            if (Cp[0]>smax) and (smax<emax):
                hrshCaseStart = (2)
                hrshCaseEnd   = -case
            if (Cp[0]<smin) and (emin<smin):
                hrshCaseStart = (-2)
                hrshCaseEnd   = -case   


        # Determining the neighbours of the start and end cells that are the horseshoe case
        startGraphNeighbours = [cellStartGraph['neighbourIndex'][ii] for ii in list(np.where(np.array(cellStartGraph['case'])==hrshCaseStart)[0])]
        endGraphNeighbours   = [cellEndGraph['neighbourIndex'][ii] for ii in list(np.where(np.array(cellEndGraph['case'])==hrshCaseEnd)[0])]

        # If there is no neighbour trim back to the edge of the cell
        if (len(startGraphNeighbours)==0) or (len(endGraphNeighbours)==0):
            if abs(case) == 2:
                self.triplet['cy'].iloc[1] = np.clip(self.triplet.iloc[1]['cy'],vmin,vmax)
            if abs(case) == 4:
                self.triplet['cx'].iloc[1] = np.clip(self.triplet.iloc[1]['cx'],vmin,vmax)        
            return

        
        if abs(hrshCaseStart) == abs(hrshCaseEnd):
            for sGN in startGraphNeighbours:
                for eGN in endGraphNeighbours:
                    if (np.array(self.neighbour_graph.loc[sGN,'neighbourIndex'])==eGN).any() and (np.array(self.neighbour_graph.loc[eGN,'neighbourIndex'])==sGN).any():
                        sGNGraph = self.neighbour_graph.loc[sGN]
                        eGNGraph = self.neighbour_graph.loc[eGN]


                        try:

                            Crp1 = np.array(cellStartGraph['neighbourCrossingPoints'])[np.where(np.array(cellStartGraph['neighbourIndex']) == sGN)[0][0],:]
                            Crp2 = np.array(sGNGraph['neighbourCrossingPoints'])[np.where(np.array(sGNGraph['neighbourIndex']) == eGN)[0][0],:]
                            Crp3 = np.array(eGNGraph['neighbourCrossingPoints'])[np.where(np.array(eGNGraph['neighbourIndex']) == cellEndGraph.name)[0][0],:]
                        
                        except:
                            if abs(case) == 2:
                                self.triplet['cy'].iloc[1] = np.clip(self.triplet.iloc[1]['cy'],vmin,vmax)
                            if abs(case) == 4:
                                self.triplet['cx'].iloc[1] = np.clip(self.triplet.iloc[1]['cx'],vmin,vmax)        
                            return

                        # Trimminng back if the cell is worse off
                        if ('SIC' in sGNGraph) and ('SIC' not in cellStartGraph) and (sGNGraph['SIC'] - cellStartGraph['SIC'])>=20:
                            if abs(case) == 2:
                                self.triplet['cy'].iloc[1] = np.clip(self.triplet.iloc[1]['cy'],vmin,vmax)
                            if abs(case) == 4:
                                self.triplet['cx'].iloc[1] = np.clip(self.triplet.iloc[1]['cx'],vmin,vmax)        
                            return

                        # Updating the origional crossing point
                        self.triplet['cx'].iloc[1]      = Crp1[0]
                        self.triplet['cy'].iloc[1]      = Crp1[1]
                        self.triplet['cellEnd'].iloc[1] = copy.deepcopy(sGNGraph)
                        self.triplet['case'].iloc[1]    = self.triplet['cellStart'].iloc[1]['case'][np.where(np.array(self.triplet['cellStart'].iloc[1]['neighbourIndex'])==sGNGraph.name)[0][0]]

                        # Crossing Point 2
                        Pcrp2 = pd.Series(name=self.triplet.iloc[1].name+1,dtype='object')
                        Pcrp2['cx']        = Crp2[0]
                        Pcrp2['cy']        = Crp2[1]
                        Pcrp2['cellStart'] = copy.deepcopy(sGNGraph)
                        Pcrp2['cellEnd']   = copy.deepcopy(eGNGraph)
                        Pcrp2['case']      = Pcrp2['cellStart']['case'][np.where(np.array(Pcrp2['cellStart']['neighbourIndex'])==Pcrp2['cellEnd'].name)[0][0]]

                        Pcrp3 = pd.Series(name=self.triplet.iloc[1].name+2,dtype='object')
                        Pcrp3['cx']        = Crp3[0]
                        Pcrp3['cy']        = Crp3[1]
                        Pcrp3['cellStart'] = copy.deepcopy(eGNGraph)
                        Pcrp3['cellEnd']   = copy.deepcopy(cellEndGraph)
                        Pcrp3['case']      = Pcrp3['cellStart']['case'][np.where(np.array(Pcrp3['cellStart']['neighbourIndex'])==Pcrp3['cellEnd'].name)[0][0]]
                        

                        self.CrossingDF = pd.concat([self.CrossingDF,Pcrp2.to_frame().transpose(),Pcrp3.to_frame().transpose()], sort=True).sort_index().reset_index(drop=True) #self.CrossingDF.append([Pcrp2,Pcrp3],sort=True).sort_index().reset_index(drop=True)

                        self.CrossingDF.index = np.arange(int(self.CrossingDF.index.min()),int(self.CrossingDF.index.max()*1e3 + 1e3),int(1e3))

                        break

            if abs(case) == 2:
                self.triplet['cy'].iloc[1] = np.clip(self.triplet.iloc[1]['cy'],vmin,vmax)
            if abs(case) == 4:
                self.triplet['cx'].iloc[1] = np.clip(self.triplet.iloc[1]['cx'],vmin,vmax)        
            return
        else:
            for sGN in startGraphNeighbours:
                for eGN in endGraphNeighbours:
                    if (np.array(sGN==eGN).any()):
                        
                        NeighGraph = self.neighbour_graph.loc[sGN]            
                        
                        
                        try:
                            Crp1 = np.array(cellStartGraph['neighbourCrossingPoints'])[np.where(np.array(cellStartGraph['neighbourIndex']) == sGN)[0][0],:]
                            Crp2 = np.array(NeighGraph['neighbourCrossingPoints'])[np.where(np.array(NeighGraph['neighbourIndex']) == cellEndGraph.name)[0][0],:]

                        except:
                            if abs(case) == 2:
                                self.triplet['cy'].iloc[1] = np.clip(self.triplet.iloc[1]['cy'],vmin,vmax)
                            if abs(case) == 4:
                                self.triplet['cx'].iloc[1] = np.clip(self.triplet.iloc[1]['cx'],vmin,vmax)        
                            return


                        # Trimminng back if the cell is worse off
                        if ('SIC' in NeighGraph) and ('SIC' not in cellStartGraph) and (NeighGraph['SIC'] - cellStartGraph['SIC'])>=20:
                            if abs(case) == 2:
                                self.triplet['cy'].iloc[1] = np.clip(self.triplet.iloc[1]['cy'],vmin,vmax)
                            if abs(case) == 4:
                                self.triplet['cx'].iloc[1] = np.clip(self.triplet.iloc[1]['cx'],vmin,vmax)        
                            return


                        # Updating the origional crossing point
                        self.triplet['cx'].iloc[1]      = Crp1[0]
                        self.triplet['cy'].iloc[1]      = Crp1[1]
                        self.triplet['cellEnd'].iloc[1] = copy.deepcopy(NeighGraph)
                        self.triplet['case'].iloc[1]    = self.triplet['cellStart'].iloc[1]['case'][np.where(np.array(self.triplet['cellStart'].iloc[1]['neighbourIndex'])==NeighGraph.name)[0][0]]

                        Pcrp2 = pd.Series(name=self.triplet.iloc[1].name+2,dtype='object')
                        Pcrp2['cx']        = Crp2[0]
                        Pcrp2['cy']        = Crp2[1]
                        Pcrp2['cellStart'] = copy.deepcopy(NeighGraph)
                        Pcrp2['cellEnd']   = copy.deepcopy(cellEndGraph)
                        Pcrp2['case']      = Pcrp2['cellStart']['case'][np.where(np.array(Pcrp2['cellStart']['neighbourIndex'])==Pcrp2['cellEnd'].name)[0][0]]
                        
                        self.CrossingDF = pd.concat([self.CrossingDF,Pcrp2.to_frame().transpose()], sort=True).sort_index().reset_index(drop=True)#self.CrossingDF.append([Pcrp2],sort=True).sort_index().reset_index(drop=True)
                        
                        self.CrossingDF.index = np.arange(int(self.CrossingDF.index.min()),int(self.CrossingDF.index.max()*1e3 + 1e3),int(1e3))

                        break

            if abs(case) == 2:
                self.triplet['cy'].iloc[1] = np.clip(self.triplet.iloc[1]['cy'],vmin,vmax)
            if abs(case) == 4:
                self.triplet['cx'].iloc[1] = np.clip(self.triplet.iloc[1]['cx'],vmin,vmax)        
            return
                

    def _reverseCase(self):
        '''
            FILL
        '''
        # Removing Reverse Edge Type 1
        startIndex = np.array([row['cellStart'].name for idx,row in self.CrossingDF.iterrows()][1:-1])
        endIndex   = np.array([row['cellEnd'].name for idx,row in self.CrossingDF.iterrows()][1:-1] )
        boolReverseEdge  = np.logical_and((startIndex[:-1] == endIndex[1:]),(startIndex[1:] == endIndex[:-1]))
        if boolReverseEdge.any():
            indxReverseEdge = np.where(boolReverseEdge)[0]+1
            for id in indxReverseEdge:
                self.CrossingDF = self.CrossingDF.drop([self.CrossingDF.iloc[id].name]).sort_index().reset_index(drop=True)


        # Removing Reverse Edge Type 2
        startIndex = np.array([row['cellStart'].name for idx,row in self.CrossingDF.iterrows()][1:-1])
        endIndex   = np.array([row['cellEnd'].name for idx,row in self.CrossingDF.iterrows()][1:-1] )
        boolReverseEdge  = (endIndex[:-1] == endIndex[1:])
        if boolReverseEdge.any():
            indxReverseEdge = np.where(boolReverseEdge)[0]+2
            for id in indxReverseEdge:
                self.CrossingDF = self.CrossingDF.drop(self.CrossingDF.iloc[id].name).sort_index().reset_index(drop=True)

        # Removing Reverse Edge Type 3
        startIndex = np.array([row['cellStart'].name for idx,row in self.CrossingDF.iterrows()][1:-1])
        endIndex   = np.array([row['cellEnd'].name for idx,row in self.CrossingDF.iterrows()][1:-1] )
        boolReverseEdge  = (startIndex[:-1] == startIndex[1:])
        if boolReverseEdge.any():
            indxReverseEdge = np.where(boolReverseEdge)[0]+1
            for id in indxReverseEdge:
                self.CrossingDF = self.CrossingDF.drop(self.CrossingDF.iloc[id].name).sort_index().reset_index(drop=True)



        self.CrossingDF.index = np.arange(0,int(len(self.CrossingDF)*1e3),int(1e3))


    def _updateCrossingPoint(self):
        '''
            FILL
        '''

        self.org_triplet = copy.deepcopy(self.triplet) 
        self.speed_s = self._unit_speed(self.triplet.iloc[1]['cellStart']['Speed'])
        self.speed_e = self._unit_speed(self.triplet.iloc[1]['cellEnd']['Speed'])

        # ------ Case Deginitions & Dealing
        if self.debugging:
            print('===========================================================')
        if abs(self.triplet.iloc[1].case)==2:
            self._long_case()
        elif abs(self.triplet.iloc[1].case)==4:
            self._lat_case()
        elif (abs(self.triplet.iloc[1].case)==1) or (abs(self.triplet.iloc[1].case)==3):
            self._corner_case()

        if len(self.triplet) < 3:
            return




    def _traveltime_in_cell(self,xdist,ydist,U,V,S):
        '''
            FILL
        '''
        dist  = np.sqrt(xdist**2 + ydist**2)
        cval  = np.sqrt(U**2 + V**2)

        dotprod  = xdist*U + ydist*V
        diffsqrs = S**2 - cval**2

        # if (dotprod**2 + diffsqrs*(dist**2) < 0)
        if diffsqrs == 0.0:
            if dotprod == 0.0:
                return np.inf
                #raise Exception(' ')
            else:
                if ((dist**2)/(2*dotprod))  <0:
                    return np.inf
                    #raise Exception(' ')
                else:
                    traveltime = dist * dist / (2 * dotprod)
                    return traveltime

        traveltime = (np.sqrt(dotprod**2 + (dist**2)*diffsqrs) - dotprod)/diffsqrs
        if traveltime < 0:
            traveltime = np.inf
        return traveltime, dist

    def _waypoint_correction(self,source_graph,Wp,Cp):
        '''
            FILL
        '''
        m_long  = 111.321*1000
        m_lat   = 111.386*1000

        x = (Cp[0]-Wp[0])*m_long*np.cos(Wp[1]*(np.pi/180))
        y = (Cp[1]-Wp[1])*m_lat
        Su  = source_graph['Vector_x']*self.zc
        Sv  = source_graph['Vector_y']*self.zc
        Ssp = self._unit_speed(source_graph['Speed'])
        traveltime, distance = self._traveltime_in_cell(x,y,Su,Sv,Ssp)
        return traveltime, distance

    def objective_function(self):
        '''
            FILL
        '''
        TravelTime = np.zeros(len(self.CrossingDF))
        Distance   = np.zeros(len(self.CrossingDF))
        index      = np.zeros(len(self.CrossingDF))
        for ii in range(len(self.CrossingDF)-1):
            soruce_graph = self.CrossingDF.iloc[ii]['cellEnd']
            Wp = self.CrossingDF.iloc[ii][['cx','cy']].to_numpy()
            Cp = self.CrossingDF.iloc[ii+1][['cx','cy']].to_numpy()
            traveltime, distance = self._waypoint_correction(soruce_graph,Wp,Cp)
            TravelTime[ii+1] = self._unit_time(traveltime)
            Distance[ii+1]   = distance

            if ii ==0:
                index[ii] = soruce_graph.name
                index[ii+1] = soruce_graph.name
            else:
                index[ii+1] = soruce_graph.name
        return TravelTime,Distance,index




"""
In this section we will discuss the usage of the Mesh
functionality of PolarRoute.

Example:
    An example of how to run this code can be executed by running the
    following in an ipython/Jupyter Notebook::

        from RoutePlanner import Mesh

        import json
        with open('./config.json', 'r') as f:
            config = json.load(f)

        mesh = Mesh(config)
"""

import logging
import math
import numpy as np
import matplotlib.pyplot as plt

# FIXME: This is weird
import json as JSON

from matplotlib.patches import Polygon as MatplotPolygon
from polar_route.cellbox import CellBox
import polar_route.data_loaders as data_loader


class Mesh:
    """
        Attributes:
            cellboxes (list<(CellBox)>): A list of CellBox objects forming the Mesh

            neighbour_graph (dict): A graphical representation of the adjacency
                relationship between CellBoxes in the Mesh. The neighbour_graph is
                of the form

               {\n
                    <CellBox id_1>: {\n
                        "1": [id_1,...,id_n],\n
                        "2": [id_1,...,id_n],\n
                        "3": [id_1,...,id_n],\n
                        "4": [id_1,...,id_n],\n
                        "-1": [id_1,...,id_n],\n
                        "-2": [id_1,...,id_n],\n
                        "-3": [id_1,...,id_n],\n
                        "-4": [id_1,...,id_n],\n
                    },\n
                    ...,\n
                    {\n
                        <CellBox id_n>: {\n
                            ...\n
                        }\n
                    }\n
               }
    """

    def __init__(self, config, j_grid=False):
        """
            Constructs a Mesh from a given config file.

            Args:
                config (dict): config file which defines the attributes of the Mesh
                    to be constructed. config is of the form - \n
                    \n
                    {\n
                        "config": {\n
                            "Mesh_info":{\n
                                "Region": {\n
                                    "latMin": (real),\n
                                    "latMax": (real),\n
                                    "longMin": (real),\n
                                    "longMax": (real),\n
                                    "startTime": (string) 'YYYY-MM-DD',\n
                                    "endTime": (string) 'YYYY-MM-DD',\n
                                    "cellWidth": (real),\n
                                    "cellHeight" (real),\n
                                    "splitDepth" (int)\n
                                },\n
                                "Data_sources": [
                                    {
                                        "loader": (string)\n
                                        "params" (dict)\n
                                    },\n
                                    ...,\n
                                    {...}

                                ]\n,
                                "splitting_conditions": [
                                    {
                                        <value>: {
                                            "threshold": (float),\n
                                            "upperBound": (float),\n
                                            "lowerBound": (float) \n
                                        }
                                    },\n
                                    ...,\n
                                    {...}\n
                                ]
                            }\n
                        }\n
                    }\n

                j_grid (bool): True if the Mesh to be constructed should be of the same
                    format as the original Java CellGrid, to be used for regression testing.
        """
        self._config = config

        self._long_min = config['Mesh_info']['Region']['longMin']
        self._long_max = config['Mesh_info']['Region']['longMax']
        self._lat_min = config['Mesh_info']['Region']['latMin']
        self._lat_max = config['Mesh_info']['Region']['latMax']

        self._cell_width = config['Mesh_info']['Region']['cellWidth']
        self._cell_height = config['Mesh_info']['Region']['cellHeight']

        assert (self._long_max - self._long_min) % self._cell_width == 0, \
            f"""The defined longitude region <{self._long_min} :{self._long_max}>
            is not divisable by the initial cell width <{self._cell_width}>"""

        assert (self._lat_max - self._lat_min) % self._cell_height == 0, \
            f"""The defined longitude region <{self._lat_min} :{self._lat_max}>
            is not divisable by the initial cell width <{self._cell_height}>"""

        self._start_time = config['Mesh_info']['Region']['startTime']
        self._end_time = config['Mesh_info']['Region']['endTime']

        # TODO: we should be using the inheritance hierarchy to achieve this
        self._j_grid = j_grid
        if 'j_grid' in config['Mesh_info'].keys():
            logging.warning("We're using the legacy Java style cell grid")
            self._j_grid = True

        # TODO: getters / setters (look at config)
        self.cellboxes = []
        self.neighbour_graph = {}
#        self.__init()


#    def __init(self):
        logging.info("Initialising mesh...")

        logging.debug("Initialise cellBoxes")
        for lat in np.arange(self._lat_min, self._lat_max, self._cell_height):
            for long in np.arange(self._long_min, self._long_max, self._cell_width):
                cellbox = CellBox(lat, long, self._cell_width, self._cell_height,
                                  splitting_conditions=[], j_grid=self._j_grid)
                self.cellboxes.append(cellbox)

        grid_width = (self._long_max - self._long_min) / self._cell_width
        grid_height = (self._lat_max - self._lat_min) / self._cell_height

        logging.debug("Calculate initial neighbours graph")
        for cellbox in self.cellboxes:
            cellbox_indx = self.cellboxes.index(cellbox)
            neighbour_map = {1: [], 2: [], 3: [], 4: [], -1: [], -2: [], -3: [], -4: []}

            # add east neighbours to neighbour graph
            if (cellbox_indx + 1) % grid_width != 0:
                neighbour_map[2].append(cellbox_indx + 1)
                # south-east neighbours
                if cellbox_indx + grid_width < len(self.cellboxes):
                    neighbour_map[1].append(int((cellbox_indx + grid_width) + 1))
                # north-east neighbours
                if cellbox_indx - grid_width >= 0:
                    neighbour_map[3].append(int((cellbox_indx - grid_width) + 1))

            # add west neighbours to neighbour graph
            if cellbox_indx % grid_width != 0:
                neighbour_map[-2].append(cellbox_indx - 1)
                # add south-west neighbours to neighbour graph
                if cellbox_indx + grid_width < len(self.cellboxes):
                    neighbour_map[-3].append(int((cellbox_indx + grid_width) - 1))
                # add north-west neighbours to neighbour graph
                if cellbox_indx - grid_width >= 0:
                    neighbour_map[-1].append(int((cellbox_indx - grid_width) - 1))

            # add south neighbours to neighbour graph
            if cellbox_indx + grid_width < len(self.cellboxes):
                neighbour_map[-4].append(int(cellbox_indx + grid_width))

            # add north neighbours to neighbour graph
            if cellbox_indx - grid_width >= 0:
                neighbour_map[4].append(int(cellbox_indx - grid_width))

            self.neighbour_graph[cellbox_indx] = neighbour_map

            # set value output types of a cellbox
            if 'value_output_types' in self.config['Mesh_info'].keys():
                cellbox.add_value_output_type(self.config['Mesh_info']['value_output_types'])

            # set gridCoord of cellBox
            x_coord = cellbox_indx % grid_width
            y_coord = abs(math.floor(cellbox_indx / grid_width) - (grid_height - 1))
            cellbox.set_grid_coord(x_coord, y_coord)

            # set focus of cellBox
            cellbox.set_focus([])

        logging.debug("calling data_loaders:")
        # j_grids represent currents differently from other data sources, so a bispoke
        # function 'add_current_points' must be called
        if self._j_grid:
            loader_name = self.config['Mesh_info']['j_grid']['Currents']['loader']
            loader = getattr(data_loader, loader_name)
            logging.debug("J_grid using loader {}".format(loader_name))

            data_points = loader(self.config['Mesh_info']['j_grid']['Currents']['params'],
                self._long_min, self._long_max, self._lat_min, self._lat_max,
                self._start_time, self._end_time)

            self.add_current_points(data_points)

        if 'Data_sources' in self.config['Mesh_info'].keys():
            for data_source in self.config['Mesh_info']['Data_sources']:
                loader = getattr(data_loader, data_source['loader'])
                logging.debug("Using data loader {}".format(data_source['loader']))

                data_points = loader(data_source['params'],
                                     self._long_min, self._long_max, self._lat_min, self._lat_max,
                                     self._start_time, self._end_time)

                self.add_data_points(data_points)

        logging.debug("Add splitting conditions from config and split Mesh.")
        self.splitting_conditions = []
        if 'splitting' in self.config['Mesh_info'].keys():
            for splitting_condition in self.config['Mesh_info']['splitting']['splitting_conditions']:
                logging.debug("Adding condition on {} to all cellboxes".format(splitting_condition))
                logging.debug("Number of cellboxes before splitting: {}".format(len(self.cellboxes)))
                for cellbox in self.cellboxes:
                    if isinstance(cellbox, CellBox):
                        cellbox.add_splitting_condition(splitting_condition)
                        cellbox.set_minimum_datapoints(self.config['Mesh_info']['splitting']['minimum_datapoints'])
                        cellbox.set_value_fill_types(self.config['Mesh_info']['splitting']['value_fill_types'])

                logging.debug("Splitting to depth {}".format(self.config['Mesh_info']['splitting']['split_depth']))
                self.split_to_depth(self.config['Mesh_info']['splitting']['split_depth'])
                logging.debug("Number of cellboxes after splitting: {}".format(len(self.cellboxes)))

    # Functions for adding data to the Mesh

    def add_data_points(self, data_points):
        """
            takes a dataframe containing geospatial-temporal located values and assigns them to
            cellboxes within this Mesh.

            Args:
                data_points (DataFrame): a dataframe of datapoints to be added to the Mesh.
                    data_points is of the form \n
                    lat | long | (time)* | value_1 | ... | value_n
        """
        for cell_box in self.cellboxes:
            if isinstance(cell_box, CellBox):
                long_loc = data_points.loc[(data_points['long'] > cell_box.long) &
                                           (data_points['long'] <= (cell_box.long + cell_box.width))]
                lat_long_loc = long_loc.loc[(long_loc['lat'] > cell_box.lat) &
                                            (long_loc['lat'] <= (cell_box.lat + cell_box.height))]

                cell_box.add_data_points(lat_long_loc)

    # Functions for outputting the Mesh

    def get_cellboxes(self):
        """
            returns a list of dictionaries containing information about each cellbox
            in this Mesh.
            all cellboxes will include id, geometry, cx, cy, dcx, dcy

            Returns:
                cellboxes (list<dict>): a list of CellBoxes which form the Mesh.
                    CellBoxes are of the form -

                    {
                        "id": (string) ... \n
                        "geometry": (string) POLYGON(...), \n
                        "cx": (float) ..., \n
                        "cy": (float) ..., \n
                        "dcx": (float) ..., \n
                        "dcy": (float) ..., \n
                        \n
                        "value_1": (float) ..., \n
                        ..., \n
                        "value_n": (float) ... \n
                    }
        """
        return_cellboxes = []
        for cellbox in self.cellboxes:
            if isinstance(cellbox, CellBox):

                # Get json for CellBox
                cell = cellbox.to_json()
                # Append ID to CellBox
                cell['id'] = str(self.cellboxes.index(cellbox))

                return_cellboxes.append(cell)
        return return_cellboxes

    def to_json(self):
        """
            Returns this Mesh converted to string parsable as a JSON object.

            Returns:
                json (string): a string representation of the CellGird parseable as a
                    JSON object. The JSON object is of the form -

                    {
                        "config": the config used to initialize the Mesh,
                        "cellboxes": a list of CellBoxes contained within the Mesh,
                        "neighbour_graph": a graph representing the adjacency of CellBoxes
                            within the Mesh
                    }
        """
        json = dict()
        json['config'] = self.config
        json["cellboxes"] = self.get_cellboxes()
        json['neighbour_graph'] = self.neighbour_graph

        # FIXME: Eh?
        return JSON.loads(JSON.dumps(json))

    # Functions for splitting cellboxes within the Mesh

    def split_and_replace(self, cellbox):
        """
            Replaces a cellBox given by parameter 'cellBox' in this grid with
            4 smaller cellBoxes representing the four corners of the given cellBox.
            A neighbours map is then created for each of the 4 new cellBoxes
            and the neighbours map for all surrounding cell boxes is updated.

            Args:
                cellbox (CellBox): the CellBox within this Mesh to be split into
                    4 smaller CellBox objects.

        """
        split_cellboxes = cellbox.split()

        self.cellboxes += split_cellboxes

        cellbox_indx = self.cellboxes.index(cellbox)

        north_west_indx = self.cellboxes.index(split_cellboxes[0])
        north_east_indx = self.cellboxes.index(split_cellboxes[1])
        south_west_indx = self.cellboxes.index(split_cellboxes[2])
        south_east_indx = self.cellboxes.index(split_cellboxes[3])

        south_neighbour_indx = self.neighbour_graph[cellbox_indx][4]
        north_neighbour_indx = self.neighbour_graph[cellbox_indx][-4]
        east_neighbour_indx = self.neighbour_graph[cellbox_indx][2]
        west_neighbour_indx = self.neighbour_graph[cellbox_indx][-2]

        # Create neighbour map for SW split cell.
        sw_neighbour_map = {1: [north_east_indx],
                            2: [south_east_indx],
                            3: [],
                            4: [],
                            -1: self.neighbour_graph[cellbox_indx][-1],
                            -2: [],
                            -3: [],
                            -4: [north_west_indx]}

        for indx in south_neighbour_indx:
            if self.get_neighbour_case(self.cellboxes[south_west_indx], self.cellboxes[indx]) == 3:
                sw_neighbour_map[3].append(indx)
            if self.get_neighbour_case(self.cellboxes[south_west_indx], self.cellboxes[indx]) == 4:
                sw_neighbour_map[4].append(indx)
        for indx in west_neighbour_indx:
            if self.get_neighbour_case(self.cellboxes[south_west_indx], self.cellboxes[indx]) == -2:
                sw_neighbour_map[-2].append(indx)
            if self.get_neighbour_case(self.cellboxes[south_west_indx], self.cellboxes[indx]) == -3:
                sw_neighbour_map[-3].append(indx)

        self.neighbour_graph[south_west_indx] = sw_neighbour_map

        # Create neighbour map for NW split cell
        nw_neighbour_map = {1: [],
                            2: [north_east_indx],
                            3: [south_east_indx],
                            4: [south_west_indx],
                            -1: [],
                            -2: [],
                            -3: self.neighbour_graph[cellbox_indx][-3],
                            -4: []}

        for indx in north_neighbour_indx:
            if self.get_neighbour_case(self.cellboxes[north_west_indx], self.cellboxes[indx]) == -4:
                nw_neighbour_map[-4].append(indx)
            if self.get_neighbour_case(self.cellboxes[north_west_indx], self.cellboxes[indx]) == 1:
                nw_neighbour_map[1].append(indx)
        for indx in west_neighbour_indx:
            if self.get_neighbour_case(self.cellboxes[north_west_indx], self.cellboxes[indx]) == -2:
                nw_neighbour_map[-2].append(indx)
            if self.get_neighbour_case(self.cellboxes[north_west_indx], self.cellboxes[indx]) == -1:
                nw_neighbour_map[-1].append(indx)

        self.neighbour_graph[north_west_indx] = nw_neighbour_map

        # Create neighbour map for NE split cell
        ne_neighbour_map = {1: self.neighbour_graph[cellbox_indx][1],
                            2: [],
                            3: [],
                            4: [south_east_indx],
                            -1: [south_west_indx],
                            -2: [north_west_indx],
                            -3: [],
                            -4: []}

        for indx in north_neighbour_indx:
            if self.get_neighbour_case(self.cellboxes[north_east_indx], self.cellboxes[indx]) == -4:
                ne_neighbour_map[-4].append(indx)
            if self.get_neighbour_case(self.cellboxes[north_east_indx], self.cellboxes[indx]) == -3:
                ne_neighbour_map[-3].append(indx)
        for indx in east_neighbour_indx:
            if self.get_neighbour_case(self.cellboxes[north_east_indx], self.cellboxes[indx]) == 2:
                ne_neighbour_map[2].append(indx)
            if self.get_neighbour_case(self.cellboxes[north_east_indx], self.cellboxes[indx]) == 3:
                ne_neighbour_map[3].append(indx)

        self.neighbour_graph[north_east_indx] = ne_neighbour_map

        # Create neighbour map for SE split cell
        se_neighbour_map = {1: [],
                            2: [],
                            3: self.neighbour_graph[cellbox_indx][3],
                            4: [],
                            -1: [],
                            -2: [south_west_indx],
                            -3: [north_west_indx],
                            -4: [north_east_indx]}

        for indx in south_neighbour_indx:
            if self.get_neighbour_case(self.cellboxes[south_east_indx], self.cellboxes[indx]) == 4:
                se_neighbour_map[4].append(indx)
            if self.get_neighbour_case(self.cellboxes[south_east_indx], self.cellboxes[indx]) == -1:
                se_neighbour_map[-1].append(indx)
        for indx in east_neighbour_indx:
            if self.get_neighbour_case(self.cellboxes[south_east_indx], self.cellboxes[indx]) == 2:
                se_neighbour_map[2].append(indx)
            if self.get_neighbour_case(self.cellboxes[south_east_indx], self.cellboxes[indx]) == 1:
                se_neighbour_map[1].append(indx)

        self.neighbour_graph[south_east_indx] = se_neighbour_map

        # Update neighbour map of neighbours of the split box.

        # Update north neighbour map
        for indx in north_neighbour_indx:
            self.neighbour_graph[indx][4].remove(cellbox_indx)

            crossing_case = self.get_neighbour_case(self.cellboxes[indx],
                                                    self.cellboxes[north_west_indx])
            if crossing_case != 0:
                self.neighbour_graph[indx][crossing_case].append(north_west_indx)

            crossing_case = self.get_neighbour_case(self.cellboxes[indx],
                                                    self.cellboxes[north_east_indx])
            if crossing_case != 0:
                self.neighbour_graph[indx][crossing_case].append(north_east_indx)

        # Update east neighbour map
        for indx in east_neighbour_indx:
            self.neighbour_graph[indx][-2].remove(cellbox_indx)

            crossing_case = self.get_neighbour_case(self.cellboxes[indx],
                                                    self.cellboxes[north_east_indx])
            if crossing_case != 0:
                self.neighbour_graph[indx][crossing_case].append(north_east_indx)

            crossing_case = self.get_neighbour_case(self.cellboxes[indx],
                                                    self.cellboxes[south_east_indx])
            if crossing_case != 0:
                self.neighbour_graph[indx][crossing_case].append(south_east_indx)

        # Update south neighbour map
        for indx in south_neighbour_indx:
            self.neighbour_graph[indx][-4].remove(cellbox_indx)

            crossing_case = self.get_neighbour_case(self.cellboxes[indx],
                                                    self.cellboxes[south_east_indx])
            if crossing_case != 0:
                self.neighbour_graph[indx][crossing_case].append(south_east_indx)

            crossing_case = self.get_neighbour_case(self.cellboxes[indx],
                                                    self.cellboxes[south_west_indx])
            if crossing_case != 0:
                self.neighbour_graph[indx][crossing_case].append(south_west_indx)

        # Update west neighbour map
        for indx in west_neighbour_indx:
            self.neighbour_graph[indx][2].remove(cellbox_indx)

            crossing_case = self.get_neighbour_case(self.cellboxes[indx],
                                                    self.cellboxes[north_west_indx])
            if crossing_case != 0:
                self.neighbour_graph[indx][crossing_case].append(north_west_indx)

            crossing_case = self.get_neighbour_case(self.cellboxes[indx],
                                                    self.cellboxes[south_west_indx])
            if crossing_case != 0:
                self.neighbour_graph[indx][crossing_case].append(south_west_indx)

        # Update corner neighbour maps
        north_east_corner_indx = self.neighbour_graph[cellbox_indx][1]
        if len(north_east_corner_indx) > 0:
            self.neighbour_graph[north_east_corner_indx[0]][-1] = [north_east_indx]

        north_west_corner_indx = self.neighbour_graph[cellbox_indx][-3]
        if len(north_west_corner_indx) > 0:
            self.neighbour_graph[north_west_corner_indx[0]][3] = [north_west_indx]

        south_east_corner_indx = self.neighbour_graph[cellbox_indx][3]
        if len(south_east_corner_indx) > 0:
            self.neighbour_graph[south_east_corner_indx[0]][-3] = [south_east_indx]

        south_west_corner_indx = self.neighbour_graph[cellbox_indx][-1]
        if len(south_west_corner_indx) > 0:
            self.neighbour_graph[south_west_corner_indx[0]][1] = [south_west_indx]

        split_container = {"northEast": north_east_indx,
                           "northWest": north_west_indx,
                           "southEast": south_east_indx,
                           "southWest": south_west_indx}

        self.cellboxes[cellbox_indx] = split_container
        self.neighbour_graph.pop(cellbox_indx)

    def split_to_depth(self, split_depth):
        """
            splits all cellboxes in this grid until a maximum split depth
            is reached, or all cellboxes are homogeneous.

            Args:
                split_depth (int): The maximum split depth reached by any CellBox
                    within this Mesh after splitting.
        """
        for cellbox in self.cellboxes:
            if isinstance(cellbox, CellBox):
                if (cellbox.split_depth < split_depth) & (cellbox.should_split()):
                    self.split_and_replace(cellbox)

    # Functions for debugging
    def plot(self, highlight_cellboxes={}, plot_ice=True, plot_currents=False,
             plot_borders=True, paths=None, routepoints=False, waypoints=None):
        """
            creates and displays a plot for this Mesh

            To be used for debugging purposes only.
        """
        # Create plot figure
        fig, axis = plt.subplots(1, 1, figsize=(25, 11))

        fig.patch.set_facecolor('white')
        axis.set_facecolor('lightblue')

        for cellbox in self.cellboxes:
            if isinstance(cellbox, CellBox):
                # plot ice
                if plot_ice and not np.isnan(cellbox.get_value('iceArea')):
                    if self._j_grid:
                        if cellbox.get_value('iceArea') >= 0.04:
                            axis.add_patch(MatplotPolygon(cellbox.get_bounds(),
                                           closed=True, fill=True, color='white', alpha=1))
                            if cellbox.get_value('iceArea') < 0.8:
                                axis.add_patch(
                                    MatplotPolygon(cellbox.get_bounds(),
                                                   closed=True, fill=True, color='grey',
                                                   alpha=(1 - cellbox.get_value('iceArea'))))
                    else:
                        axis.add_patch(MatplotPolygon(cellbox.get_bounds(), closed=True,
                                       fill=True, color='white', alpha=cellbox.get_value('iceArea')))

                # plot land
                if self._j_grid:
                    if cellbox.land_locked:
                        axis.add_patch(MatplotPolygon(cellbox.get_bounds(),
                                                      closed=True, fill=True, facecolor='lime'))
                else:
                    if cellbox.contains_land():
                        axis.add_patch(MatplotPolygon(cellbox.get_bounds(),
                                                      closed=True, fill=True, facecolor='mediumseagreen'))

                # plot currents
                if plot_currents:
                    axis.quiver((cellbox.long + cellbox.width / 2),
                                (cellbox.lat + cellbox.height / 2),
                                cellbox.get_value('uC'), cellbox.get_value('vC'),
                                scale=1, width=0.002, color='gray')

                # plot borders
                if plot_borders:
                    axis.add_patch(MatplotPolygon(cellbox.get_bounds(),
                                                  closed=True, fill=False, edgecolor='black'))

        # plot highlighted cells
        for colour in highlight_cellboxes:
            for cellbox in highlight_cellboxes[colour]:
                axis.add_patch(MatplotPolygon(cellbox.get_bounds(),
                                              closed=True,
                                              fill=False,
                                              edgecolor=colour,
                                              linewidth=0.5 + len(list(highlight_cellboxes.keys())) -
                                              list(highlight_cellboxes.keys()).index(colour)))

        # plot paths if supplied
        if paths is not None:
            for path in paths:
                if path['Time'] == np.inf:
                    continue
                points = np.array(path['Path']['Points'])
                if routepoints:
                    axis.plot(points[:, 0], points[:, 1], linewidth=3.0, color='b')
                    axis.scatter(points[:, 0], points[:, 1], 30, zorder=99, color='b')
                else:
                    axis.plot(points[:, 0], points[:, 1], linewidth=3.0, color='b')

        if waypoints is not None:
            axis.scatter(waypoints['Long'], waypoints['Lat'], 150, marker='^', color='r')

        axis.set_xlim(self._long_min, self._long_max)
        axis.set_ylim(self._lat_min, self._lat_max)

    def get_neighbour_case(self, cellbox_a, cellbox_b):
        """
            Given two cellBoxes (cellbox_a, cellbox_b) returns a case number
            representing where the two cellBoxes are touching.

            Args:
                cellbox_a (CellBox): starting CellBox
                cellbox_b (CellBox): destination CellBox

            Returns:
                case (int): an int representing the direction of the adjacency
                    between input cellbox_a and cellbox_b. The meaning of each case
                    is as follows -

                        case 0 -> CellBoxes are not neighbours

                        case 1 -> cellbox_b is the North-East corner of cellbox_a\n
                        case 2 -> cellbox_b is East of cellbox_a\n
                        case 3 -> cellbox_b is the South-East corner of cellbox_a\n
                        case 4 -> cellbox_b is South of cellbox_a\n
                        case -1 -> cellbox_b is the South-West corner of cellbox_a\n
                        case -2 -> cellbox_b is West of cellbox_a\n
                        case -3 -> cellbox_b is the North-West corner of cellbox_a\n
                        case -4 -> cellbox_b is North of cellbox_a\n
        """

        if (cellbox_a.long + cellbox_a.width) == cellbox_b.long and (
                cellbox_a.lat + cellbox_a.height) == cellbox_b.lat:
            return 1  # North-East
        if (cellbox_a.long + cellbox_a.width == cellbox_b.long) and (
                cellbox_b.lat < (cellbox_a.lat + cellbox_a.height)) and (
                (cellbox_b.lat + cellbox_b.height) > cellbox_a.lat):
            return 2  # East
        if (cellbox_a.long + cellbox_a.width) == cellbox_b.long and (
                cellbox_a.lat == cellbox_b.lat + cellbox_b.height):
            return 3  # South-East
        if ((cellbox_b.lat + cellbox_b.height) == cellbox_a.lat) and (
                (cellbox_b.long + cellbox_b.width) > cellbox_a.long) and (
                cellbox_b.long < (cellbox_a.long + cellbox_a.width)):
            return 4  # South
        if cellbox_a.long == (cellbox_b.long + cellbox_b.width) and cellbox_a.lat == (
                cellbox_b.lat + cellbox_b.height):
            return -1  # South-West
        if (cellbox_b.long + cellbox_b.width == cellbox_a.long) and (
                cellbox_b.lat < (cellbox_a.lat + cellbox_a.height)) and (
                (cellbox_b.lat + cellbox_b.height) > cellbox_a.lat):
            return -2  # West
        if cellbox_a.long == (cellbox_b.long + cellbox_b.width) and (
                cellbox_a.lat + cellbox_a.height == cellbox_b.lat):
            return -3  # North-West
        if (cellbox_b.lat == (cellbox_a.lat + cellbox_a.height)) and (
                (cellbox_b.long + cellbox_b.width) > cellbox_a.long) and (
                cellbox_b.long < (cellbox_a.long + cellbox_a.width)):
            return -4  # North
        return 0  # Cells are not neighbours.

    def get_cellbox(self, long, lat):
        """
            Returns the CellBox which contains a point, given by parameters lat, long

            Args:
                long (long): longitude of a given point
                lat (float): latitude of given point

            Returns:
                cellbox (CellBox): the cellbox which contains the point given my parameters
                (long, lat)
        """
        selected_cell = []
        for cellbox in self.cellboxes:
            if isinstance(cellbox, CellBox):
                if cellbox.contains_point(lat, long):
                    selected_cell.append(cellbox)
        return selected_cell[0]

    # Functions used for j_grid regression testing
    def dump_mesh(self, file_location):
        """
            creates a string representation of this Mesh which
            is then saved to a file location specified by parameter
            'file_location'

            for use in j_grid regression testing
        """
        mesh_dump = ""
        for cell_box in self.cellboxes:
            if isinstance(cell_box, CellBox):
                mesh_dump += cell_box.mesh_dump()

        file = open(file_location, "w")
        file.write(mesh_dump)
        file.close()

    def dump_graph(self, file_location):
        """
            creates a string representation of the neighbour relation
            of this Mesh which is then saved to a file location#
            specified by parameter 'file_location'

            for use in j_grid regression testing
        """
        graph_dump = ""

        max_ice_area = 0.8

        for cellbox in self.cellboxes:
            if isinstance(cellbox, CellBox):
                if (not cellbox.land_locked) and cellbox.get_value('SIC') < max_ice_area:
                    graph_dump += cellbox.node_string()

                    cellbox_indx = self.cellboxes.index(cellbox)

                    # case -3 neighbours
                    nw_neighbour_indx = self.neighbour_graph[cellbox_indx][-3]
                    for neighbour in nw_neighbour_indx:
                        if (not self.cellboxes[neighbour].land_locked) and (
                                self.cellboxes[neighbour].get_value('SIC') < max_ice_area):
                            graph_dump += "," + self.cellboxes[neighbour].node_string() + ":-3"
                    # case -2 neighbours
                    w_neighbours_indx = self.neighbour_graph[cellbox_indx][-2]
                    for neighbour in w_neighbours_indx:
                        if (not self.cellboxes[neighbour].land_locked) and (
                                self.cellboxes[neighbour].get_value('SIC') < max_ice_area):
                            graph_dump += "," + self.cellboxes[neighbour].node_string() + ":-2"
                    # case -1 neighbours
                    sw_neighbours_indx = self.neighbour_graph[cellbox_indx][-1]
                    for neighbour in sw_neighbours_indx:
                        if (not self.cellboxes[neighbour].land_locked) and (
                                self.cellboxes[neighbour].get_value('SIC') < max_ice_area):
                            graph_dump += "," + self.cellboxes[neighbour].node_string() + ":-1"
                    # case -4 neighbours
                    n_neighbour_indx = self.neighbour_graph[cellbox_indx][-4]
                    for neighbour in n_neighbour_indx:
                        if (not self.cellboxes[neighbour].land_locked) and (
                            self.cellboxes[neighbour].get_value('SIC') < max_ice_area):
                            graph_dump += "," + self.cellboxes[neighbour].node_string() + ":-4"
                    # case 4 neighbours
                    s_neighbours_indx = self.neighbour_graph[cellbox_indx][4]
                    for neighbour in s_neighbours_indx:
                        if (not self.cellboxes[neighbour].land_locked) and (
                                self.cellboxes[neighbour].get_value('SIC') < max_ice_area):
                            graph_dump += "," + self.cellboxes[neighbour].node_string() + ":4"
                    # case 1 neighbours
                    ne_neighbour_indx = self.neighbour_graph[cellbox_indx][1]
                    for neighbour in ne_neighbour_indx:
                        if (not self.cellboxes[neighbour].land_locked) and (
                                self.cellboxes[neighbour].get_value('SIC') < max_ice_area):
                            graph_dump += "," + self.cellboxes[neighbour].node_string() + ":1"
                    # case 2 neighbours
                    e_neighbour_indx = self.neighbour_graph[cellbox_indx][2]
                    for neighbour in e_neighbour_indx:
                        if (not self.cellboxes[neighbour].land_locked) and (
                                self.cellboxes[neighbour].get_value('SIC') < max_ice_area):
                            graph_dump += "," + self.cellboxes[neighbour].node_string() + ":2"
                    # case 3 neighbours
                    se_neighbour_indx = self.neighbour_graph[cellbox_indx][3]
                    for neighbour in se_neighbour_indx:
                        if (not self.cellboxes[neighbour].land_locked) and (
                                self.cellboxes[neighbour].get_value('SIC') < max_ice_area):
                            graph_dump += "," + self.cellboxes[neighbour].node_string() + ":3"

                    graph_dump += "\n"

        file = open(file_location, "w")
        file.write(graph_dump)
        file.close()

    def add_current_points(self, current_points):
        """
            Takes a dataframe containing current points and assigns
            then to cellboxes within the Mesh
        """
        for cellbox in self.cellboxes:
            long_loc = current_points.loc[(current_points['long'] > cellbox.long) & (
                current_points['long'] <= (cellbox.long + cellbox.width))]
            lat_long_loc = long_loc.loc[(long_loc['lat'] > cellbox.lat) & (
                long_loc['lat'] <= (cellbox.lat + cellbox.height))]

            cellbox.add_current_points(lat_long_loc)

            # find data point closest to centre to determin land
            def closest_point(df, lat, long):
                n_lat_df = df.iloc[(df['lat'] - lat).abs().argsort()[:15]]
                n_long_df = n_lat_df.iloc[(n_lat_df['long'] - long).abs().argsort()[:1]]
                return n_long_df

            point = closest_point(lat_long_loc, cellbox.getcy(), cellbox.getcx())
            if np.isnan(point['uC'].mean()) or np.isnan(point['vC'].mean()):
                cellbox.land_locked = True

            #cellbox.set_land()

    def cellbox_by_node_string(self, node_string):
        """
            given a node string specified by parameter 'node_string'
            returns a cellbox object which that node string identifies

            used for debugging of j_grids.
        """
        for cellbox in self.cellboxes:
            if isinstance(cellbox, CellBox):
                if cellbox.node_string() == node_string:
                    return cellbox

    @property
    def config(self):
        return self._config